""" BlueSky aircraft performance calculations."""
from math import *
import numpy as np
import bluesky as bs
from bluesky.tools.aero import ft, g0, a0, T0, rho0, gamma1, gamma2,  beta, R, \
    kts, lbs, inch, sqft, fpm, vtas2cas, vcas2tas
from bluesky.traffic.performance.perfbase import PerfBase
from bluesky.traffic.performance.legacy.performance import esf, phases, calclimits, PHASE
from bluesky import settings

from bluesky.traffic.performance.legacy.coeff_bs import CoeffBS

# Register settings defaults
settings.set_variable_defaults(performance_dt=1.0, verbose=False)
coeffBS = CoeffBS()


class Legacy(PerfBase):
    def __init__(self):
        super().__init__()
        self.warned  = False    # Flag: Did we warn for default perf parameters yet?
        self.warned2 = False    # Flag: Use of piston engine aircraft?

        # prepare for coefficient readin
        coeffBS.coeff()

        with self.settrafarrays():
            # index of aircraft types in library
            self.coeffidxlist = np.array([])

            # geometry and weight
            self.mass         = np.array([]) # Mass [kg]
            self.Sref         = np.array([]) # Wing surface area [m^2]

            # reference velocities
            self.refma        = np.array([]) # reference Mach
            self.refcas       = np.array([]) # reference CAS
            self.gr_acc       = np.array([]) # ground acceleration
            self.gr_dec       = np.array([]) # ground deceleration

            # limits
            self.vm_to        = np.array([]) # min takeoff spd (w/o mass, density)
            self.vm_ld        = np.array([]) # min landing spd (w/o mass, density)
            self.vmto         = np.array([]) # min TO spd
            self.vmic         = np.array([]) # min. IC speed
            self.vmcr         = np.array([]) # min cruise spd
            self.vmap         = np.array([]) # min approach speed
            self.vmld         = np.array([]) # min landing spd
            self.vmo          = np.array([]) # max CAS
            self.mmo          = np.array([]) # max Mach

            self.hmaxact      = np.array([]) # max. altitude
            self.maxthr       = np.array([]) # maximum thrust

            # aerodynamics
            self.CD0          = np.array([]) # parasite drag coefficient
            self.k            = np.array([]) # induced drag factor
            self.clmaxcr      = np.array([]) # max. cruise lift coefficient
            self.qS           = np.array([]) # Dynamic air pressure [Pa]

            # engines
            self.n_eng        = np.array([]) # Number of engines
            self.etype        = np.array([]) # jet /turboprop

            # jet engines:
            self.rated_thrust = np.array([]) # rated thrust (all engines)
            self.SFC = np.array([]) # specific fuel consumption in cruise
            self.fuelflow = np.array([]) # fuel flow
            self.ffto = np.array([]) # fuel flow takeoff
            self.ffcl = np.array([]) # fuel flow climb
            self.ffcr = np.array([]) # fuel flow cruise
            self.ffid = np.array([]) # fuel flow idle
            self.ffap = np.array([]) # fuel flow approach

            # turboprop engines
            self.P = np.array([]) # avaliable power at takeoff conditions
            self.PSFC_TO = np.array([]) # specific fuel consumption takeoff
            self.PSFC_CR = np.array([]) # specific fuel consumption cruise

            self.thrust = np.array([]) # Thrust
            self.thrust_pilot = np.array([])   # thrust required for pilot settings
            self.D = np.array([]) # Drag
            self.ESF = np.array([]) # Energy share factor according to EUROCONTROL

            # flight phase
            self.phase = np.array([]) # flight phase
            self.bank = np.array([]) # bank angle
            self.post_flight = np.array([]) # check for ground mode:
                                              #taxi prior of after flight
            self.pf_flag = np.array([])

            self.engines = []           # avaliable engine type per aircraft type

            # limit settings
            self.limspd      = np.array([])  # limit speed
            self.limspd_flag = np.array([], dtype=bool)  # flag for limit spd - we have to test for max and min
            self.limalt      = np.array([])  # limit altitude
            self.limalt_flag = np.array([])  # A need to limit altitude has been detected
            self.limvs       = np.array([])  # limit vertical speed due to thrust limitation
            self.limvs_flag  = np.array([])  # A need to limit V/S detected
        self.eta = 0.8          # propeller efficiency according to Raymer
        self.thrust_settings = np.array([1., 0.85, 0.07, 0.3 ]) # Thrust settings per flight phase according to ICAO

    def create(self, n=1):
        super().create(n)
        """CREATE NEW AIRCRAFT"""
        actypes = bs.traf.type[-n:]
        coeffidx = []

        for actype in actypes:
            if actype in coeffBS.atype:
                coeffidx.append(coeffBS.atype.index(actype))
            else:
                coeffidx.append(0)
                if not settings.verbose:
                    if not self.warned:
                        print("Aircraft is using default B747-400 performance.")
                        self.warned = True
                else:
                    print("Flight " + bs.traf.id[-1] + " has an unknown aircraft type, " + actype + ", BlueSky then uses default B747-400 performance.")
        coeffidx = np.array(coeffidx)

        # note: coefficients are initialized in SI units

        self.coeffidxlist[-n:]      = coeffidx
        self.mass[-n:]              = coeffBS.MTOW[coeffidx] # aircraft weight
        self.Sref[-n:]              = coeffBS.Sref[coeffidx] # wing surface reference area
        self.etype[-n:]             = coeffBS.etype[coeffidx] # engine type of current aircraft
        self.engines[-n:]           = [coeffBS.engines[c] for c in coeffidx]

        # speeds
        self.refma[-n:]             = coeffBS.cr_Ma[coeffidx] # nominal cruise Mach at 35000 ft
        self.refcas[-n:]            = vtas2cas(coeffBS.cr_spd[coeffidx], 35000*ft) # nominal cruise CAS
        self.gr_acc[-n:]            = coeffBS.gr_acc[coeffidx] # ground acceleration
        self.gr_dec[-n:]            = coeffBS.gr_dec[coeffidx] # ground acceleration

        # limits
        self.vm_to[-n:]             = coeffBS.vmto[coeffidx]
        self.vm_ld[-n:]             = coeffBS.vmld[coeffidx]
        self.mmo[-n:]               = coeffBS.max_Ma[coeffidx] # maximum Mach
        self.vmo[-n:]               = coeffBS.max_spd[coeffidx] # maximum CAS
        self.hmaxact[-n:]           = coeffBS.max_alt[coeffidx] # maximum altitude
        # self.vmto/vmic/vmcr/vmap/vmld/vmin are initialised as 0 by super.create

        # aerodynamics
        self.CD0[-n:]               = coeffBS.CD0[coeffidx]  # parasite drag coefficient
        self.k[-n:]                 = coeffBS.k[coeffidx]    # induced drag factor
        self.clmaxcr[-n:]           = coeffBS.clmax_cr[coeffidx]   # max. cruise lift coefficient
        self.ESF[-n:]               = 1.
        # self.D/qS are initialised as 0 by super.create

        # flight phase
        self.pf_flag[-n:]           = 1
        # self.phase/bank/post_flight are initialised as 0 by super.create

        # engines
        self.n_eng[-n:]              = coeffBS.n_eng[coeffidx] # Number of engines
        turboprops = self.etype[-n:] == 2

        propidx = []
        jetidx  = []

        for engine in self.engines[-n:]:
            # engine[0]: default to first engine in engine list for this aircraft
            if engine[0] in coeffBS.propenlist:
                propidx.append(coeffBS.propenlist.index(engine[0]))
            else:
                propidx.append(0)
            if engine[0] in coeffBS.jetenlist:
                jetidx.append(coeffBS.jetenlist.index(engine[0]))
            else:
                jetidx.append(0)

        propidx=np.array(propidx)
        jetidx =np.array(jetidx)
        # Make two index lists of the engine type, assuming jet and prop. In the end, choose which one to use

        self.P[-n:]         = np.where(turboprops, coeffBS.P[propidx]*self.n_eng[-n:]      , 1.)
        self.PSFC_TO[-n:]   = np.where(turboprops, coeffBS.PSFC_TO[propidx]*self.n_eng[-n:], 1.)
        self.PSFC_CR[-n:]   = np.where(turboprops, coeffBS.PSFC_CR[propidx]*self.n_eng[-n:], 1.)

        self.rated_thrust[-n:]      = np.where(turboprops, 1. , coeffBS.rated_thrust[jetidx]*coeffBS.n_eng[coeffidx])  # rated thrust (all engines)
        self.thrust[-n:]       = np.where(turboprops, 1. , coeffBS.rated_thrust[jetidx]*coeffBS.n_eng[coeffidx])  # initialize thrust with rated thrust
        self.thrust_pilot[-n:] = np.where(turboprops, 1. , coeffBS.rated_thrust[jetidx]*coeffBS.n_eng[coeffidx])  # initialize thrust with rated thrust
        self.maxthr[-n:]    = np.where(turboprops, 1. , coeffBS.rated_thrust[jetidx]*coeffBS.n_eng[coeffidx]*1.2)  # maximum thrust - initialize with 1.2*rThr
        self.SFC[-n:]       = np.where(turboprops, 1. , coeffBS.SFC[jetidx] )
        self.ffto[-n:]      = np.where(turboprops, 1. , coeffBS.ffto[jetidx]*coeffBS.n_eng[coeffidx]) / 60.0
        self.ffcl[-n:]      = np.where(turboprops, 1. , coeffBS.ffcl[jetidx]*coeffBS.n_eng[coeffidx]) / 60.0
        self.ffcr[-n:]      = np.where(turboprops, 1. , coeffBS.ffcr[jetidx]*coeffBS.n_eng[coeffidx]) / 60.0
        self.ffid[-n:]      = np.where(turboprops, 1. , coeffBS.ffid[jetidx]*coeffBS.n_eng[coeffidx]) / 60.0
        self.ffap[-n:]      = np.where(turboprops, 1. , coeffBS.ffap[jetidx]*coeffBS.n_eng[coeffidx]) / 60.0

    def update(self, dt):
        ''' Periodic update function for performance calculations. '''
        swbada = False # no-bada version
        delalt = bs.traf.selalt - bs.traf.alt
        # allocate aircraft to their flight phase
        self.phase, self.bank = phases(bs.traf.alt, bs.traf.gs, delalt,
            bs.traf.cas, self.vmto, self.vmic, self.vmap, self.vmcr, self.vmld,
            bs.traf.ap.bankdef, bs.traf.bphase, bs.traf.swhdgsel,swbada)

        # AERODYNAMICS
        # compute CL: CL = 2*m*g/(VTAS^2*rho*S)
        self.qS = 0.5*bs.traf.rho*np.maximum(1.,bs.traf.tas)*np.maximum(1.,bs.traf.tas)*self.Sref

        cl = self.mass*g0/(self.qS*np.cos(self.bank))*(self.phase!=6)+ 0.*(self.phase==6)

        # scaling factors for CD0 and CDi during flight phases according to FAA (2005): SAGE, V. 1.5, Technical Manual

        # For takeoff (phase = 6) drag is assumed equal to the takeoff phase
        CD0f = (self.phase==1)*(self.etype==1)*coeffBS.d_CD0j[0] + \
               (self.phase==2)*(self.etype==1)*coeffBS.d_CD0j[1]  + \
               (self.phase==3)*(self.etype==1)*coeffBS.d_CD0j[2] + \
               (self.phase==4)*(self.etype==1)*coeffBS.d_CD0j[3] + \
               (self.phase==5)*(self.etype==1)*(bs.traf.alt>=450.0)*coeffBS.d_CD0j[4] + \
               (self.phase==5)*(self.etype==1)*(bs.traf.alt<450.0)*coeffBS.d_CD0j[5] + \
               (self.phase==6)*(self.etype==1)*coeffBS.d_CD0j[0] + \
               (self.phase==1)*(self.etype==2)*coeffBS.d_CD0t[0] + \
               (self.phase==2)*(self.etype==2)*coeffBS.d_CD0t[1]  + \
               (self.phase==3)*(self.etype==2)*coeffBS.d_CD0t[2] + \
               (self.phase==4)*(self.etype==2)*coeffBS.d_CD0t[3]
                   # (self.phase==5)*(self.etype==2)*(self.alt>=450)*coeffBS.d_CD0t[4] + \
                   # (self.phase==5)*(self.etype==2)*(self.alt<450)*coeffBS.d_CD0t[5]

        # For takeoff (phase = 6) induced drag is assumed equal to the takeoff phase
        kf =   (self.phase==1)*(self.etype==1)*coeffBS.d_kj[0] + \
               (self.phase==2)*(self.etype==1)*coeffBS.d_kj[1]  + \
               (self.phase==3)*(self.etype==1)*coeffBS.d_kj[2] + \
               (self.phase==4)*(self.etype==1)*coeffBS.d_kj[3] + \
               (self.phase==5)*(self.etype==1)*(bs.traf.alt>=450)*coeffBS.d_kj[4] + \
               (self.phase==5)*(self.etype==1)*(bs.traf.alt<450)*coeffBS.d_kj[5] + \
               (self.phase==6)*(self.etype==1)*coeffBS.d_kj[0] + \
               (self.phase==1)*(self.etype==2)*coeffBS.d_kt[0] + \
               (self.phase==2)*(self.etype==2)*coeffBS.d_kt[1]  + \
               (self.phase==3)*(self.etype==2)*coeffBS.d_kt[2] + \
               (self.phase==4)*(self.etype==2)*coeffBS.d_kt[3] + \
               (self.phase==5)*(self.etype==2)*(bs.traf.alt>=450)*coeffBS.d_kt[4] + \
               (self.phase==5)*(self.etype==2)*(bs.traf.alt<450)*coeffBS.d_kt[5]


        # drag coefficient
        cd = self.CD0*CD0f + self.k*kf*(cl*cl)

        # compute drag: CD = CD0 + CDi * CL^2 and D = rho/2*VTAS^2*CD*S
        self.D = cd*self.qS
        # energy share factor and crossover altitude
        epsalt = np.array([0.001]*bs.traf.ntraf)
        climb = np.array(delalt > epsalt)
        descent = np.array(delalt< -epsalt)

        # energy share factor
        delspd = bs.traf.aporasas.tas - bs.traf.tas
        selmach = bs.traf.selspd < 2.0
        self.ESF = esf(bs.traf.alt, bs.traf.M, climb, descent, delspd, selmach)

        # determine thrust
        self.thrust = (((bs.traf.vs*self.mass*g0)/(self.ESF*np.maximum(bs.traf.eps, bs.traf.tas))) + self.D)
        # determine thrust required to fulfill requests from pilot
        # self.thrust_pilot = (((bs.traf.aporasas.vs*self.mass*g0)/(self.ESF*np.maximum(bs.traf.eps, bs.traf.aporasas.tas))) + self.D)
        self.thrust_pilot = (((bs.traf.ap.vs*self.mass*g0)/(self.ESF*np.maximum(bs.traf.eps, bs.traf.aporasas.tas))) + self.D)

        # maximum thrust jet (Bruenig et al., p. 66):
        mt_jet = self.rated_thrust*(bs.traf.rho/rho0)**0.75

        # maximum thrust prop (Raymer, p.36):
        mt_prop = self.P*self.eta/np.maximum(bs.traf.eps, bs.traf.tas)

        # merge
        self.maxthr = mt_jet*(self.etype==1) + mt_prop*(self.etype==2)

        # Fuel Flow

        # jet aircraft
        # ratio current thrust/rated thrust
        pThr = self.thrust/self.rated_thrust
        # fuel flow is assumed to be proportional to thrust(Torenbeek, p.62).
        #For ground operations, idle thrust is used
        # cruise thrust is approximately equal to approach thrust
        ff_jet = ((pThr*self.ffto)*(self.phase!=6)*(self.phase!=3)+ \
        self.ffid*(self.phase==6) + self.ffap*(self.phase==3) )*(self.etype==1)
        # print "FFJET",  (pThr*self.ffto)*(self.phase!=6)*(self.phase!=3), self.ffid*(self.phase==6), self.ffap*(self.phase==3)
        # print "FFJET", ff_jet

        # turboprop aircraft
        # to be refined - f(spd)
        # CRUISE-ALTITUDE!!!
        # above cruise altitude: PSFC_CR
        PSFC = (((self.PSFC_CR - self.PSFC_TO) / 20000.0)*bs.traf.alt + self.PSFC_TO)*(bs.traf.alt<20.000) + \
                self.PSFC_CR*(bs.traf.alt >= 20.000)

        TSFC = PSFC*bs.traf.tas/(550.0*self.eta)

        # formula p.36 Raymer is missing here!
        ff_prop = self.thrust*TSFC*(self.etype==2)


        # combine
        self.fuelflow = np.maximum(0.0, ff_jet + ff_prop)

        # update mass
        self.mass -= self.fuelflow * dt

        # print bs.traf.id, self.phase, bs.traf.alt/ft, bs.traf.tas/kts, bs.traf.cas/kts, bs.traf.M,  \
        # self.thrust, self.D, self.ff,  cl, cd, bs.traf.vs/fpm, self.ESF,self.atrans, self.maxthr, \
        # self.vmto/kts, self.vmic/kts ,self.vmcr/kts, self.vmap/kts, self.vmld/kts, \
        # CD0f, kf, self.hmaxact


        # for aircraft on the runway and taxiways we need to know, whether they
        # are prior or after their flight
        self.post_flight = np.where(descent, True, self.post_flight)

        # when landing, we would like to stop the aircraft.
        bs.traf.aporasas.tas = np.where((bs.traf.alt <0.5)*(self.post_flight)*self.pf_flag, 0.0, bs.traf.aporasas.tas)
        # the impulse for reducing the speed to 0 should only be given once,
        # otherwise taxiing will be impossible afterwards
        self.pf_flag = np.where ((bs.traf.alt <0.5)*(self.post_flight), False, self.pf_flag)

        # define acceleration: aircraft taxiing and taking off use ground acceleration,
        # landing aircraft use ground deceleration, others use standard acceleration
        self.axmax = ((self.phase == PHASE['IC']) + (self.phase == PHASE['CR']) + (self.phase == PHASE['AP']) + (self.phase == PHASE['LD'])) * 0.5 \
            + ((self.phase == PHASE['TO']) + (self.phase == PHASE['GD'])*(1-self.post_flight)) * self.gr_acc  \
            + (self.phase == PHASE['GD']) * self.post_flight * self.gr_dec

    def limits(self, intent_v, intent_vs, intent_h, ax):
        """Flight envelope""" # Connect this with function limits in performance.py

        # combine minimum speeds and flight phases. Phases initial climb, cruise
        # and approach use the same CLmax and thus the same function for Vmin
        self.vmto = self.vm_to*np.sqrt(self.mass/bs.traf.rho)
        self.vmic = np.sqrt(2*self.mass*g0/(bs.traf.rho*self.clmaxcr*self.Sref))
        self.vmcr = self.vmic
        self.vmap = self.vmic
        self.vmld = self.vm_ld*np.sqrt(self.mass/bs.traf.rho)

        # summarize and convert to cas
        # note: aircraft on ground may be pushed back
        self.vmin = (self.phase==1)*vtas2cas(self.vmto, bs.traf.alt) + \
                        ((self.phase==2) + (self.phase==3) + (self.phase==4))*vtas2cas(self.vmcr, bs.traf.alt) + \
                            (self.phase==5)*vtas2cas(self.vmld, bs.traf.alt) + (self.phase==6)*-10.0


        # forwarding to tools
        self.limspd, self.limspd_flag, self.limalt, \
            self.limalt_flag, self.limvs, self.limvs_flag  =  calclimits(
                vtas2cas(intent_v, bs.traf.alt), bs.traf.gs, \
                self.vmto, self.vmin, self.vmo, self.mmo, \
                bs.traf.M, bs.traf.alt, self.hmaxact, \
                intent_h, intent_vs, self.maxthr, \
                self.thrust_pilot, self.D, bs.traf.tas, \
                self.mass, self.ESF, self.phase)

        # Update desired sates with values within the flight envelope
        # When CAS is limited, it needs to be converted to TAS as only this TAS is used later on!
        allowed_tas = np.where(self.limspd_flag, vcas2tas(self.limspd, bs.traf.alt), intent_v)

        # Autopilot selected altitude [m]
        allowed_alt = np.where(self.limalt_flag, self.limalt, intent_h)

        # Autopilot selected vertical speed (V/S)
        allowed_vs = np.where(self.limvs_flag, self.limvs, intent_vs)

        return allowed_tas, allowed_vs, allowed_alt

    def engchange(self, idx, engid=None):
        """change of engines - for jet aircraft only!"""
        if not engid:
            disptxt = "available engine types:\n" + '\n'.join(self.engines[idx]) + \
                      "\nChange engine with ENG acid engine_id"
            return False, disptxt
        engidx = self.engines[idx].index(engid)
        self.jetengidx = coeffBS.jetenlist.index(coeffBS.engines[idx][engidx])

        # exchange engine parameters

        self.rated_thrust[idx]   = coeffBS.rated_thrust[self.jetengidx]*coeffBS.n_eng[idx] # rated thrust (all engines)
        self.thrust[idx]    = coeffBS.rated_thrust[self.jetengidx]*coeffBS.n_eng[idx] # initialize thrust with rated thrust
        self.maxthr[idx] = coeffBS.rated_thrust[self.jetengidx]*coeffBS.n_eng[idx] # maximum thrust - initialize with 1.2*rThr
        self.SFC[idx]    = coeffBS.SFC[self.jetengidx]
        self.fuelflow[idx] = 0. # neutral initialisation
        self.ffto[idx]   = coeffBS.ffto[self.jetengidx]*coeffBS.n_eng[idx] / 60.0
        self.ffcl[idx]   = coeffBS.ffcl[self.jetengidx]*coeffBS.n_eng[idx] / 60.0
        self.ffcr[idx]   = coeffBS.ffcr[self.jetengidx]*coeffBS.n_eng[idx] / 60.0
        self.ffid[idx]   = coeffBS.ffid[self.jetengidx]*coeffBS.n_eng[idx] / 60.0
        self.ffap[idx]   = coeffBS.ffap[self.jetengidx]*coeffBS.n_eng[idx] / 60.0
        return



''' The BlueSky command stack.

    The stack parses all text-based commands in the simulation.
'''
from bluesky.stack.stackbase import stack, forward, sender, get_scenname, get_scendata, set_scendata
from bluesky.stack.cmdparser import command, commandgroup, append_commands, \
    remove_commands, get_commands
from bluesky.stack.argparser import ArgumentError


def process(ext_cmds):
    ''' Dummy process that will get replaced by sim or client implementation during init. '''
    pass


def init(mode='client'):
    if mode[:3] == 'sim':
        import bluesky.stack.simstack as simstack
        from bluesky.stack.importer import Importer
        simstack.init()
        globals()['process'] = simstack.process
    else:
        import bluesky.stack.clientstack as clientstack
        clientstack.init()
        globals()['process'] = clientstack.process



''' BlueSky functions for geographical calculations. '''
from bluesky import settings
from bluesky.tools.geo._geo import nm, magdec, initdecl_data, magdeccmd, kwikpos


# Register settings defaults
settings.set_variable_defaults(prefer_compiled=False)
if getattr(settings, 'prefer_compiled'):
    try:
        from bluesky.tools.geo._cgeo import (
            rwgs84,
            rwgs84_matrix,
            qdrdist,
            qdrdist_matrix,
            latlondist,
            latlondist_matrix,
            wgsg,
            qdrpos,
            kwikdist,
            kwikdist_matrix,
            kwikqdrdist,
            kwikqdrdist_matrix
        )
        print('Using compiled geo functions')
    except ImportError:
        from bluesky.tools.geo._geo import (rwgs84, rwgs84_matrix, qdrdist, 
                                        qdrdist_matrix, latlondist, latlondist_matrix,
                                        wgsg, qdrpos, kwikdist, kwikdist_matrix, 
                                        kwikqdrdist, kwikqdrdist_matrix)
        print('Could not load compiled geo functions, Using Python-based geo functions instead')
else:
    from bluesky.tools.geo._geo import (rwgs84, rwgs84_matrix, qdrdist, 
                                        qdrdist_matrix, latlondist, latlondist_matrix,
                                        wgsg, qdrpos, kwikdist, kwikdist_matrix, 
                                        kwikqdrdist, kwikqdrdist_matrix)
    print('Using Python-based geo functions')



"""
This is a decoder of ABS-D date from Mode-S receiver. The inputs
    of most functions are the Hexdecial strings.

Created by  : Junzi Sun (TU Delft)
Date        : March 2015
"""

import math

MODES_CHECKSUM_TABLE = [
    0x3935ea, 0x1c9af5, 0xf1b77e, 0x78dbbf,
    0xc397db, 0x9e31e9, 0xb0e2f0, 0x587178,
    0x2c38bc, 0x161c5e, 0x0b0e2f, 0xfa7d13,
    0x82c48d, 0xbe9842, 0x5f4c21, 0xd05c14,
    0x682e0a, 0x341705, 0xe5f186, 0x72f8c3,
    0xc68665, 0x9cb936, 0x4e5c9b, 0xd8d449,
    0x939020, 0x49c810, 0x24e408, 0x127204,
    0x093902, 0x049c81, 0xfdb444, 0x7eda22,
    0x3f6d11, 0xe04c8c, 0x702646, 0x381323,
    0xe3f395, 0x8e03ce, 0x4701e7, 0xdc7af7,
    0x91c77f, 0xb719bb, 0xa476d9, 0xadc168,
    0x56e0b4, 0x2b705a, 0x15b82d, 0xf52612,
    0x7a9309, 0xc2b380, 0x6159c0, 0x30ace0,
    0x185670, 0x0c2b38, 0x06159c, 0x030ace,
    0x018567, 0xff38b7, 0x80665f, 0xbfc92b,
    0xa01e91, 0xaff54c, 0x57faa6, 0x2bfd53,
    0xea04ad, 0x8af852, 0x457c29, 0xdd4410,
    0x6ea208, 0x375104, 0x1ba882, 0x0dd441,
    0xf91024, 0x7c8812, 0x3e4409, 0xe0d800,
    0x706c00, 0x383600, 0x1c1b00, 0x0e0d80,
    0x0706c0, 0x038360, 0x01c1b0, 0x00e0d8,
    0x00706c, 0x003836, 0x001c1b, 0xfff409,
    0x000000, 0x000000, 0x000000, 0x000000,
    0x000000, 0x000000, 0x000000, 0x000000,
    0x000000, 0x000000, 0x000000, 0x000000,
    0x000000, 0x000000, 0x000000, 0x000000,
    0x000000, 0x000000, 0x000000, 0x000000,
    0x000000, 0x000000, 0x000000, 0x000000
]


def hex2bin(hexstr):
    """Convert a hexdecimal string to binary string, with zero fillings. """
    length = len(hexstr) * 4
    msgbin = bin(int(hexstr, 16))[2:]
    while ((len(msgbin)) < length):
        msgbin = '0' + msgbin
    return msgbin


def bin2int(binstr):
    return int(binstr, 2)


def hex2int(hexstr):
    return int(hexstr, 16)


def checksum(msg):
    if len(msg) == 28:
        offset = 0
    elif len(msg) == 14:
        offset = 112-56
    else:
        # raise exception
        return False

    msgbin = hex2bin(msg)
    checksum = int(msg[22:28], 16)

    crc = 0
    for i in range(len(msgbin)):
        if int(msgbin[i]):
            crc ^= MODES_CHECKSUM_TABLE[i+offset]

    if crc == checksum:
        return True
    else:
        return False


def get_df(msg):
    """Decode Downlink Format vaule, bits 1 to 5."""
    msgbin = hex2bin(msg)
    return bin2int(msgbin[0: 5])


def get_ca(msg):
    """Decode CA vaule, bits: 6 to 8."""
    msgbin = hex2bin(msg)
    return bin2int(msgbin[5:8])


def get_icao_addr(msg):
    """Get the ICAO 24 bits address, bytes 3 to 8. """
    return msg[2:8]


def get_tc(msg):
    """Get Type Code, bits 33 to 37 """
    msgbin = hex2bin(msg)
    return bin2int(msgbin[32:37])


def get_oe_flag(msg):
    """Check the odd/even flag. Bit 54, 0 for even, 1 for odd."""
    msgbin = hex2bin(msg)
    return msgbin[53]


def get_alt(msg):
    """Calculate the altitude from the message. Bit 41 to 52, Q-bit at 48"""
    msgbin = hex2bin(msg)
    q = msgbin[47]
    if q:
        n = bin2int(msgbin[40:47]+msgbin[48:52])
        alt = n * 25 - 1000
        return alt
    else:
        return None


def get_cprlat(msg):
    msgbin = hex2bin(msg)
    return bin2int(msgbin[54:71])


def get_cprlon(msg):
    msgbin = hex2bin(msg)
    return bin2int(msgbin[71:88])


def get_position(msg0, msg1, t0, t1):
    cprlat0 = get_cprlat(msg0)
    cprlat1 = get_cprlat(msg1)
    cprlon0 = get_cprlon(msg0)
    cprlon1 = get_cprlon(msg1)
    return cpr2position(cprlat0, cprlat1, cprlon0, cprlon1, t0, t1)


def cpr2position(cprlat0, cprlat1, cprlon0, cprlon1, t0, t1):
    '''
     This algorithm comes from:
     http://www.lll.lu/~edward/edward/adsb/DecodingADSBposition.html.

     131072 is 2^17 since CPR latitude and longitude are encoded in 17 bits.
    '''

    cprlat_even = cprlat0 / 131072.0
    cprlat_odd = cprlat1 / 131072.0
    cprlon_even = cprlon0 / 131072.0
    cprlon_odd = cprlon1 / 131072.0

    air_d_lat_even = 360.0 / 60
    air_d_lat_odd = 360.0 / 59

    # compute latitude index 'j'
    j = int(59 * cprlat_even - 60 * cprlat_odd + 0.5)

    lat_even = float(air_d_lat_even * (j % 60 + cprlat_even))
    lat_odd = float(air_d_lat_odd * (j % 59 + cprlat_odd))

    if lat_even >= 270:
        lat_even = lat_even - 360

    if lat_odd >= 270:
        lat_odd = lat_odd - 360

    # check if both are in the same latidude zone, exit if not
    if cprNL(lat_even) != cprNL(lat_odd):
        return None

    # compute ni, longitude index m, and longitude
    if (t0 > t1):
        ni = cprN(lat_even, 0)
        m = math.floor(cprlon_even * (cprNL(lat_even)-1)
                       - cprlon_odd * cprNL(lat_even) + 0.5)
        lon = (360.0 / ni) * (m % ni + cprlon_even)
        lat = lat_even
    else:
        ni = cprN(lat_odd, 1)
        m = math.floor(cprlon_even * (cprNL(lat_odd)-1)
                       - cprlon_odd * cprNL(lat_odd) + 0.5)
        lon = (360.0 / ni) * (m % ni + cprlon_odd)
        lat = lat_odd

    if lon > 180:
        lon = lon - 360

    return [lat, lon]


def get_speed_heading(msg):
    """Calculate the speed and heading."""
    msgbin = hex2bin(msg)

    v_ew_dir = bin2int(msgbin[45])
    v_ew = bin2int(msgbin[46:56])       # east-west velocity

    v_ns_dir = bin2int(msgbin[56])
    v_ns = bin2int(msgbin[57:67])       # north-south velocity

    v_ew = -1*v_ew if v_ew_dir else v_ew
    v_ns = -1*v_ns if v_ns_dir else v_ns

    # vr       = bin2int(msgbin[68:77])       # vertical rate
    # vr_dir   = bin2int(msgbin[77])

    speed = math.sqrt(v_ns*v_ns + v_ew*v_ew)    # unit in kts

    heading = math.atan2(v_ew, v_ns)
    heading = heading * 360.0 / (2 * math.pi)   # convert to degrees
    heading = heading if heading >= 0 else heading + 360     # no negative val
    return [speed, heading]


def get_callsign(msg):
    """Decode aircraft identification, aka. Callsign"""

    chars = '#ABCDEFGHIJKLMNOPQRSTUVWXYZ#####_###############0123456789######'
    msgbin = hex2bin(msg)
    csbin = msgbin[40:96]

    cs = ''
    cs += chars[bin2int(csbin[0:6])]
    cs += chars[bin2int(csbin[6:12])]
    cs += chars[bin2int(csbin[12:18])]
    cs += chars[bin2int(csbin[18:24])]
    cs += chars[bin2int(csbin[24:30])]
    cs += chars[bin2int(csbin[30:36])]
    cs += chars[bin2int(csbin[36:42])]
    cs += chars[bin2int(csbin[42:48])]

    # clean string, remove spaces and marks, if any.
    cs = cs.replace('_', '')
    cs = cs.replace('#', '')
    return cs


def cprN(lat, is_odd):
    nl = cprNL(lat) - is_odd
    return nl if nl > 1 else 1


def cprNL(lat):
    try:
        nz = 60
        a = 1 - math.cos(math.pi * 2 / nz)
        b = math.cos(math.pi / 180.0 * abs(lat)) ** 2
        nl = 2 * math.pi / (math.acos(1 - a/b))
        return int(nl)
    except:
        # happens when latitude is +/-90 degree
        return 1



''' Main simulation-side stack functions. '''
import math
from pathlib import Path
import traceback
import bluesky as bs
from bluesky.stack.stackbase import Stack, stack, checkscen, forward
from bluesky.stack.cmdparser import Command, command
from bluesky.stack.basecmds import initbasecmds
from bluesky.stack import recorder
from bluesky.stack import argparser, ArgumentError
from bluesky import settings


# Register settings defaults
settings.set_variable_defaults(start_location="EHAM", scenario_path="scenario")

# List of TMX commands not yet implemented in BlueSky
tmxlist = ("BGPASAS", "DFFLEVEL", "FFLEVEL", "FILTCONF", "FILTTRED", "FILTTAMB",
           "GRAB", "HDGREF", "MOVIE", "NAVDB", "PREDASAS", "RENAME", "RETYPE",
           "SWNLRPASAS", "TRAFRECDT", "TRAFLOGDT", "TREACT", "WINDGRID")


def init():
    """ Initialization of the default stack commands. This function is called
        at the initialization of the main simulation object."""

    # Initialise base commands
    initbasecmds()

    # Pan to initial location
    stack("PAN " + settings.start_location)
    stack("ZOOM 0.4")


def reset():
    """ Reset the stack. """

    Stack.reset()

    # Close recording file and reset scenario recording settings
    recorder.reset()


def process(ext_cmds=None):
    ''' Sim-side stack processing. '''
    # If no commands are passed, check for commands in scenario file
    if ext_cmds is None:
        checkscen()

    # Process stack of commands
    for cmdline in Stack.commands(ext_cmds):
        success = True
        echotext = ''
        echoflags = bs.BS_OK

        # Get first argument from command line and check if it's a command
        cmd, argstring = argparser.getnextarg(cmdline)
        cmdu = cmd.upper()
        cmdobj = Command.cmddict.get(cmdu)

        # If no function is found for 'cmd', check if cmd is actually an aircraft id
        if not cmdobj and cmdu in bs.traf.id:
            cmd, argstring = argparser.getnextarg(argstring)
            argstring = cmdu + " " + argstring
            # When no other args are parsed, command is POS
            cmdu = cmd.upper() if cmd else 'POS'
            cmdobj = Command.cmddict.get(cmdu)

        # Proceed if a command object was found
        if cmdobj:
            try:
                # Call the command, passing the argument string
                success, echotext = cmdobj(argstring)
                if not success:
                    if not argstring:
                        echotext = echotext or cmdobj.brieftext()
                    else:
                        echoflags = bs.BS_FUNERR
                        echotext = f'Syntax error: {echotext or cmdobj.brieftext()}'

            except ArgumentError as e:
                success = False
                echoflags = bs.BS_ARGERR
                header = '' if not argstring else e.args[0] if e.args else 'Argument error.'
                echotext = f'{header}\nUsage:\n{cmdobj.brieftext()}'
            except Exception as e:
                echoflags = bs.BS_FUNERR
                header = '' if not argstring else e.args[0] if e.args else 'Function error.'
                echotext = f'Error calling function implementation of {cmdu}: {header}\n' + \
                    'Traceback printed to terminal.'
                traceback.print_exc()

        # ----------------------------------------------------------------------
        # ZOOM command (or use ++++  or --  to zoom in or out)
        # ----------------------------------------------------------------------
        elif cmdu[0] in ("+", "=", "-"):
            # = equals + (same key)
            nplus = cmdu.count("+") + cmdu.count("=")
            nmin = cmdu.count("-")
            bs.scr.zoom(math.sqrt(2) ** (nplus - nmin), absolute=False)
            cmdu = 'ZOOM'

        # -------------------------------------------------------------------
        # Command not found
        # -------------------------------------------------------------------
        elif Stack.sender_id is None:
            # Command came from scenario file: assume it's a gui/client command and send it on
            forward()
        else:
            success = False
            echoflags = bs.BS_CMDERR
            if not argstring:
                echotext = f'Unknown command or aircraft: {cmd}'
            else:
                echotext = f'Unknown command: {cmd}'

        # Recording of actual validated commands
        if success:
            recorder.savecmd(cmdu, cmdline)
        elif not Stack.sender_id:
            echotext = f'{cmdline}\n{echotext}'

        # Always return on command
        if echotext:
            bs.scr.echo(echotext, echoflags)

    # Clear the processed commands
    if ext_cmds is None:
        Stack.clear()


def readscn(fname):
    ''' Read a scenario file. '''
    if not fname:
        return
    # Ensure .scn suffix and specify path if necessary
    fname = Path(fname).with_suffix('.scn')
    if not fname.is_absolute():
        fname = bs.resource(settings.scenario_path) / fname

    with open(fname, "r") as fscen:
        prevline = ''
        for line in fscen:
            line = line.strip()
            # Skip emtpy lines and comments
            if len(line) < 10 or line[0] == "#":
                continue
            line = prevline + line

            # Check for line continuation
            if line[-1] == '\\':
                prevline = f'{line[:-1].strip()} '
                continue
            prevline = ''

            # Try reading timestamp and command
            try:
                line = line.split("#")[0].strip()

                icmdline = line.index(">")
                tstamp = line[:icmdline]
                ttxt = tstamp.strip().split(":")
                ihr = int(ttxt[0]) * 3600.0
                imin = int(ttxt[1]) * 60.0
                xsec = float(ttxt[2])
                cmdtime = ihr + imin + xsec

                yield (cmdtime, line[icmdline + 1:].strip("\n"))
            except (ValueError, IndexError):
                # nice try, we will just ignore this syntax error
                if not (len(line.strip()) > 0 and line.strip()[0] == "#"):
                    print("except this:" + line)


@command(aliases=('CALL',), brief="PCALL filename [REL/ABS/args]")
def pcall(fname, *pcall_arglst):
    """ PCALL: Import another scenario file into the current scenario.

        Arguments:
        - fname: The filename of the scenario file to import
        - REL/ABS: indicate whether timestamps in imported scenario file should
          be considered relative to current simulation time (the default),
          or absolute. (optional argument)
        - pcall_arglist: optional additional arguments to pass to the
          imported scenario. Replaces %0, %1, ... in the scenario file.
    """
    # Check for a/c id as first argument (use case: procedure files)
    # CALL KL204 myproc should have effect as if: CALL myproc KL204
    if pcall_arglst and fname in bs.traf.id:
        acid = fname
        fname = pcall_arglst[0]
        pcall_arglst = [acid] + list(pcall_arglst[1:])

    # Check for relative or absolute time
    isrelative = True  # default relative to the time of call
    if pcall_arglst and pcall_arglst[0] in ('ABS', 'REL'):
        isrelative = pcall_arglst[0] == 'ABS'
        pcall_arglst = pcall_arglst[1:]

    try:
        merge(readscn(fname), *pcall_arglst, isrelative=isrelative)

    except FileNotFoundError as e:
        return False, f"PCALL: File not found'{e.filename}'"


def merge(source, *args, isrelative=True):
    ''' Merge scenario commands from source to current scenario.'''

    # If timestamps in file should be interpreted as relative we need to add
    # the current simtime to every timestamp
    t_offset = bs.sim.simt if isrelative else 0.0

    # Read the scenario file
    insidx = 0
    instime = bs.sim.simt

    # All commands with timestamps at the current sim time or earlier should be called immediately
    callnow = []
    for (cmdtime, cmdline) in source:

        # Time offset correction
        cmdtime += t_offset

        # Replace %0, %1 with pcall_arglst[0], pcall_arglst[1], etc.
        if args:
            for i, argtxt in enumerate(args):
                cmdline = cmdline.replace(f"%{i}", argtxt)

        if cmdtime <= bs.sim.simt:
            callnow.append((cmdline, None))
        elif not Stack.scentime or cmdtime >= Stack.scentime[-1]:
            Stack.scentime.append(cmdtime)
            Stack.scencmd.append(cmdline)
        else:
            if cmdtime > instime:
                insidx, instime = next(
                    ((j, t) for j, t in enumerate(Stack.scentime) if t >= cmdtime),
                    (len(Stack.scentime), Stack.scentime[-1]),
                )
            Stack.scentime.insert(insidx, cmdtime)
            Stack.scencmd.insert(insidx, cmdline)
            insidx += 1

    # execute any commands that are already due
    if callnow:
        process(callnow)


@command(aliases=('LOAD', 'OPEN'))
def ic(filename : 'string' = ''):
    ''' IC: Load a scenario file (initial condition).

        Arguments:
        - filename: The filename of the scenario to load. Call IC IC
          to load previous scenario again. '''

    # Get the filename of new scenario
    if not filename:
        filename = bs.scr.show_file_dialog()
        if not filename:
            # Only PyGame returns a filename from the dialog here
            return

    # reset sim always
    bs.sim.reset()

    # Clean up filename
    filename = Path(filename)

    # Reset sim and open new scenario file
    try:
        for (cmdtime, cmd) in readscn(filename):
            Stack.scentime.append(cmdtime)
            Stack.scencmd.append(cmd)
        Stack.scenname = filename.stem

        # Remember this filename in IC.scn in scenario folder
        with open(bs.resource(settings.scenario_path) / "ic.scn", "w") as keepicfile:
            keepicfile.write(
                "# This file is used by BlueSky to save the last used scenario file\n"
            )
            keepicfile.write(
                "# So in the console type 'IC IC' to restart the previously used scenario file\n"
            )
            keepicfile.write(f"00:00:00.00>IC {filename}\n")

        return True, f"IC: Opened {filename}"
    except FileNotFoundError:
        return False, f"IC: File not found: {filename}"


@command(aliases=('SCEN',))
def scenario(name: 'string'):
    """ SCENARIO sets the scenario name for the current simulation.

        Arguments:
        - name: The name to give the scenario """
    Stack.scenname = name
    return True, "Starting scenario " + name


@command
def schedule(time: 'time', cmdline: 'string'):
    """ SCHEDULE a stack command at a specific simulation time.

        Arguments:
        - time: the time at which the command should be executed
        - cmdline: the command line to be executed """
    # Get index of first scentime greater than 'time' as insert position
    idx = next((i for i, t in enumerate(Stack.scentime) if t > time), len(Stack.scentime))
    Stack.scentime.insert(idx, time)
    Stack.scencmd.insert(idx, cmdline)
    return True


@command
def delay(time: 'time', cmdline: 'string'):
    """ DELAY a stack command until a specific simulation time.

        Arguments:
        - time: the time with which the command should be delayed
        - cmdline: the command line to be executed after the delay """
    # Get index of first scentime greater than 'time' as insert position
    time += bs.sim.simt
    idx = next((i for i, t in enumerate(Stack.scentime) if t > time), len(Stack.scentime))
    Stack.scentime.insert(idx, time)
    Stack.scencmd.insert(idx, cmdline)
    return True


@command(name='HELP', aliases=('?',))
def showhelp(cmd:'txt'='', subcmd:'txt'=''):
    """ HELP: Display general help text or help text for a specific command,
        or dump command reference in file when command is >filename.

        Arguments:
        - cmd: Argument can refer to:
            - Command name to display help for. 
            - Call HELP >filename to generate a CSV file with help text for all commands.
    """

    # Check if help is asked for a specific command
    cmdobj = Command.cmddict.get(cmd or 'HELP')
    if cmdobj:
        return True, cmdobj.helptext(subcmd)

    # Write command reference to tab-delimited text file
    if cmd[0] == ">":
        # Get filename
        if len(cmd) > 1:
            fname = "./docs/" + cmd[1:]
        else:
            fname = "./docs/bluesky-commands.txt"

        # Get unique set of commands
        cmdobjs = set(Command.cmddict.values())
        table = []  # for alphabetical sort use a table

        # Get info for all commands
        for obj in cmdobjs:
            fname = obj.callback.__name__.replace("<", "").replace(">", "")
            args = ','.join((str(p) for p in obj.parsers))
            syn = ','.join(obj.aliases)
            line = f'{obj.name}\t{obj.help}\t{obj.brief}\t{args}\t{fname}\t{syn}'
            table.append(line)

        # Sort & write table
        table.sort()
        with open(fname, "w") as f:
            # Header of first table
            f.write("Command\tDescription\tUsage\tArgument types\tFunction\tSynonyms\n")
            f.write('\n'.join(table))
        return True, "Writing command reference in " + fname

    return False, "HELP: Unknown command: " + cmd


@command
def makedoc():
    ''' MAKEDOC: Make markdown templates for all stack functions
        that don't have a doc page yet.
    '''
    tmp = Path('tmp')
    if not tmp.is_dir():
        tmp.mkdir()
    # Get unique set of commands
    cmdobjs = set(Command.cmddict.values())
    for o in cmdobjs:
        if not bs.resource(f"html/{o.name}.html").is_file():
            with open(tmp / f"{o.name.lower()}.md", "w") as f:
                f.write(
                    f"# {o.name}: {o.name.capitalize()}\n"
                    + o.help
                    + "\n\n"
                    + "**Usage:**\n\n"
                    + f"    {o.brief}\n\n"
                    + "**Arguments:**\n\n"
                )
                if not o.params:
                    f.write("This command has no arguments.\n\n")
                else:
                    f.write(
                        "|Name|Type|Optional|Description\n"
                        + "|--------|------|---|---------------------------------------------------\n"
                    )
                    for arg in o.params:
                        f.write(str(arg).replace(':', '|') + f" |{arg.hasdefault()}|\n")
                f.write("\n[[Back to command reference.|Command Reference]]\n")


@command(aliases=tmxlist)
def tmx(*args):
    ''' Stub function for TMX commands that aren't available yet in BlueSky. '''
    return True, 'This TMX command has not (yet) been implemented in BlueSky.'



""" Basic implementation of a calculator.
    These functions are imported for use in eval-function for the calculator
    functionality in the BlueSky console."""

from math import *      # Make all math function available for calculator

# Allow using bluesky variables in expression

import bluesky as bs

# Cater for use of some geo functions
from .geo import latlondist as dist
from .geo import rwgs84 as wgs84
from .geo import *
from .aero import *

# And for conversion lat/lon formats
from .misc import latlon2txt,lat2txt,lon2txt

# Some special functions for calculator:
# Degree variant of sin,cos,tan:
def sind(x):
    return sin(radians(x))
def cosd(x):
    return cos(radians(x))
def tand(x):
    return tan(radians(x))

# Conversion from degrees/minutes/seconds
def rad(d,m,s):
    return radians(float(d)+float(m)/60.+float(s)/3600.)
def deg(d,m,s):
    return float(d)+float(m)/60.+float(s)/3600.

# Short-hand for sqrt()
def v(x):
    return sqrt(x)

#Def qdr function to return value without dist
def qdr(lata,lona,latb,lonb):
    return qdrdist(lata,lona,latb,lonb)[0]

def calculator(txt):
    # Simple calculator which can use math functions
    try:
        x = eval(txt) ## First try direct
    except:
        try:
            expr = txt.lower().replace("^","**") # Allow ^ for power
            x = eval(expr)
        except:
            return False,"Error in calculating "+txt+"\n"+ \
            "Use math functions, pi, e and/or sind() cosd()"\
                +" tand() deg(d,m,s) rad(d,m,s)\n"      +   \
            " or geo functions: dist,qdr,qdrdist,qdrpos,rwgs84," \
             +"kwikdist,latlondist,lat2txt,lon2txt,latlon2txt"
    return True, str(x)



import numpy as np
from bluesky.tools import aero
from bluesky.traffic.performance.openap import phase as ph


def compute_max_thr_ratio(phase, bpr, v, h, vs, thr0):
    """Computer the dynamic thrust based on engine bypass-ratio, static maximum
    thrust, aircraft true airspeed, and aircraft altitude

    Args:
        phase (int or 1D-array): phase of flight, option: phase.[NA, TO, IC, CL,
            CR, DE, FA, LD, GD]
        bpr (int or 1D-array): engine bypass ratio
        v (int or 1D-array): aircraft true airspeed
        h (int or 1D-array): aircraft altitude

    Returns:
        int or 1D-array: thust in N
    """

    n = len(phase)

    # ---- thrust ratio at takeoff ----
    ratio_takeoff = tr_takeoff(bpr, v, h)

    # ---- thrust ratio in flight ----
    ratio_inflight = inflight(v, h, vs, thr0)

    # thrust ratio array
    #   LD and GN assume ZERO thrust
    tr = np.ones(n) * ratio_inflight
    tr = np.where(phase == ph.GD, ratio_takeoff, tr)

    return tr


def tr_takeoff(bpr, v, h):
    """Compute thrust ration at take-off"""
    G0 = 0.0606 * bpr + 0.6337
    Mach = aero.vtas2mach(v, h)
    P0 = aero.p0
    P = aero.vpressure(h)
    PP = P / P0

    A = -0.4327 * PP ** 2 + 1.3855 * PP + 0.0472
    Z = 0.9106 * PP ** 3 - 1.7736 * PP ** 2 + 1.8697 * PP
    X = 0.1377 * PP ** 3 - 0.4374 * PP ** 2 + 1.3003 * PP

    ratio = (
        A
        - 0.377 * (1 + bpr) / np.sqrt((1 + 0.82 * bpr) * G0) * Z * Mach
        + (0.23 + 0.19 * np.sqrt(bpr)) * X * Mach ** 2
    )

    return ratio


def inflight(v, h, vs, thr0):
    """Compute thrust ration for inflight"""

    def dfunc(mratio):
        d = -0.4204 * mratio + 1.0824
        return d

    def nfunc(roc):
        n = 2.667e-05 * roc + 0.8633
        return n

    def mfunc(vratio, roc):
        m = -1.2043e-1 * vratio - 8.8889e-9 * roc ** 2 + 2.4444e-5 * roc + 4.7379e-1
        return m

    roc = np.abs(np.asarray(vs / aero.fpm))
    v = np.where(v < 10, 10, v)

    mach = aero.vtas2mach(v, h)
    vcas = aero.vtas2cas(v, h)

    p = aero.vpressure(h)
    p10 = aero.vpressure(10000 * aero.ft)
    p35 = aero.vpressure(35000 * aero.ft)

    # approximate thrust at top of climb (REF 2)
    F35 = (200 + 0.2 * thr0 / 4.448) * 4.448
    mach_ref = 0.8
    vcas_ref = aero.vmach2cas(mach_ref, 35000 * aero.ft)

    # segment 3: alt > 35000:
    d = dfunc(mach / mach_ref)
    b = (mach / mach_ref) ** (-0.11)
    ratio_seg3 = d * np.log(p / p35) + b

    # segment 2: 10000 < alt <= 35000:
    a = (vcas / vcas_ref) ** (-0.1)
    n = nfunc(roc)
    ratio_seg2 = a * (p / p35) ** (-0.355 * (vcas / vcas_ref) + n)

    # segment 1: alt <= 10000:
    F10 = F35 * a * (p10 / p35) ** (-0.355 * (vcas / vcas_ref) + n)
    m = mfunc(vcas / vcas_ref, roc)
    ratio_seg1 = m * (p / p35) + (F10 / F35 - m * (p10 / p35))

    ratio = np.where(
        h > 35000 * aero.ft,
        ratio_seg3,
        np.where(h > 10000 * aero.ft, ratio_seg2, ratio_seg1),
    )

    # convert to maximum static thrust ratio
    ratio_F0 = ratio * F35 / thr0

    return ratio_F0


def compute_eng_ff_coeff(ffidl, ffapp, ffco, ffto):
    """Compute fuel flow based on engine icao fuel flow model

    Args:
        thrust_ratio (1D-array): thrust ratio between 0 and 1
        n_engines (1D-array): number of engines on the aircraft
        ff_idl (1D-array): fuel flow - idle thrust
        ff_app (1D-array): fuel flow - approach
        ff_co (1D-array): fuel flow - climb out
        ff_to (1D-array): fuel flow - takeoff

    Returns:
        list of coeff: [a, b, c], fuel flow calc: ax^2 + bx + c
    """

    # standard fuel flow at test thrust ratios
    y = [0, ffidl, ffapp, ffco, ffto]
    x = [0, 0.07, 0.3, 0.85, 1.0]

    a, b, c = np.polyfit(x, y, 2)

    return a, b, c



""" This module defines a set of standard aerodynamic functions and constants."""
# Vectorized versions of aero conversion routines
from math import *
import numpy as np

from bluesky import settings


settings.set_variable_defaults(casmach_threshold=2.0)
# International standard atmpshere only up to 72000 ft / 22 km

#
# Constants Aeronautics
#
kts = 0.514444              # m/s  of 1 knot
ft  = 0.3048                # m    of 1 foot
fpm = ft/60.                # feet per minute
inch = 0.0254               # m    of 1 inch
sqft = 0.09290304           # 1sqft
nm  = 1852.                 # m    of 1 nautical mile
lbs = 0.453592              # kg   of 1 pound mass
g0  = 9.80665               # m/s2    Sea level gravity constant
R   = 287.05287             # Used in wikipedia table: checked with 11000 m
p0 = 101325.                # Pa     Sea level pressure ISA
rho0 = 1.225                # kg/m3  Sea level density ISA
T0   = 288.15               # K   Sea level temperature ISA
Tstrat = 216.65             # K Stratosphere temperature (until alt=22km)
gamma = 1.40                # cp/cv: adiabatic index for air
gamma1 =  0.2               # (gamma-1)/2 for air
gamma2 = 3.5                # gamma/(gamma-1) for air
beta = -0.0065              # [K/m] ISA temp gradient below tropopause
Rearth = 6371000.           # m  Average earth radius
a0  = np.sqrt(gamma*R*T0)   # sea level speed of sound ISA
casmach_thr = settings.casmach_threshold # Threshold below which speeds should
                            # be considered as Mach numbers in casormach* functions


def casmachthr(threshold:float=None):
    """ CASMACHTHR threshold

        Set a threshold below which speeds should be considered as Mach numbers
        in CRE(ATE), ADDWPT, and SPD commands. Set to zero if speeds should
        never be considered as Mach number (e.g., when simulating drones).

        Argument:
        - threshold: CAS speed threshold [m/s]
    """
    if threshold is None:
        return True, f'CASMACHTHR: The current CAS/Mach threshold is {casmach_thr} m/s ({casmach_thr / kts} kts'

    globals()['casmach_thr'] = threshold
    return True, f'CASMACHTHR: Set CAS/Mach threshold to {threshold}'


#
# Functions for aeronautics in this module
#  - physical quantities always in SI units
#  - lat,lon,course and heading in degrees
#
#  International Standard Atmosphere up to 22 km
#
#   p,rho,T = vatmos(h)    # atmos as function of geopotential altitude h [m]
#   a = vvsound(h)         # speed of sound [m/s] as function of h[m]
#   p = vpressure(h)       # calls atmos but retruns only pressure [Pa]
#   T = vtemperature(h)    # calculates temperature [K] (saves time rel to atmos)
#   rho = vdensity(h)      # calls atmos but retruns only pressure [Pa]
#
#  Speed conversion at altitude h[m] in ISA:
#
# M   = vtas2mach(tas,h)  # true airspeed (tas) to mach number conversion
# tas = vmach2tas(M,h)    # true airspeed (tas) to mach number conversion
# tas = veas2tas(eas,h)   # equivalent airspeed to true airspeed, h in [m]
# eas = vtas2eas(tas,h)   # true airspeed to equivent airspeed, h in [m]
# tas = vcas2tas(cas,h)   # cas  to tas conversion both m/s, h in [m]
# cas = vtas2cas(tas,h)   # tas to cas conversion both m/s, h in [m]
# cas = vmach2cas(M,h)    # Mach to cas conversion cas in m/s, h in [m]
# M   = vcas2mach(cas,h)   # cas to mach copnversion cas in m/s, h in [m]

# Atmosphere up to 22 km (72178 ft)


# ------------------------------------------------------------------------------
# Vectorized aero functions
# ------------------------------------------------------------------------------
def vatmos(h):
    """ Calculate atmospheric pressure, density, and temperature for a given altitude.

        Arguments:
        - h: Altitude [m]

        Returns:
        - p: Pressure [Pa]
        - rho: Density [kg / m3]
        - T: Temperature [K]
    """
    # Temp
    T = vtemp(h)

    # Density
    rhotrop = 1.225 * (T / 288.15)**4.256848030018761
    dhstrat = np.maximum(0., h - 11000.)
    rho     = rhotrop * np.exp(-dhstrat / 6341.552161)  # = *g0/(287.05*216.65))

    # Pressure
    p = rho * R * T

    return p, rho, T


def vtemp(h):
    """ Calculate atmospheric temperature for a given altitude.

        Arguments:
        - h: Altitude [m]

        Returns:
        - T: Temperature [K]
    """
    T = np.maximum(288.15 - 0.0065 * h, Tstrat)
    return T


# Atmos wrappings:
def vpressure(h):
    """ Calculate atmospheric pressure for a given altitude.

        Arguments:
        - h: Altitude [m]

        Returns:
        - p: Pressure [Pa]
    """
    p, _, _ = vatmos(h)
    return p


def vdensity(h):
    """ Calculate atmospheric density for a given altitude.

        Arguments:
        - h: Altitude [m]

        Returns:
        - rho: Density [kg / m3]
    """
    _, r, _ = vatmos(h)
    return r


def vvsound(h):
    """ Calculate the speed of sound for a given altitude.

        Arguments:
        - h: Altitude [m]

        Returns:
        - a: Speed of sound [m/s]
    """
    T = vtemp(h)
    a = np.sqrt(gamma * R * T)
    return a


# ---------Speed conversions---h in [m]------------------
def vtas2mach(tas, h):
    """ True airspeed (tas) to mach number conversion for numpy arrays.

        Arguments:
        - tas: True airspeed [m/s]
        - h: Altitude [m]

        Returns:
        - M: Mach number [-]
    """
    a = vvsound(h)
    mach = tas / a
    return mach


def vmach2tas(mach, h):
    """ Mach number to True airspeed (tas) conversion for numpy arrays.

        Arguments:
        - mach: Mach number [-]
        - h: Altitude [m]

        Returns:
        - tas: True airspeed [m/s]
    """
    a = vvsound(h)
    tas = mach * a
    return tas


def veas2tas(eas, h):
    """ Equivalent airspeed to true airspeed conversion for numpy arrays.

        Arguments:
        - eas: Equivalent airspeed [m/s]
        - h: Altitude [m]

        Returns:
        - tas: True airspeed [m/s]
    """
    rho = vdensity(h)
    tas = eas * np.sqrt(rho0 / rho)
    return tas


def vtas2eas(tas, h):
    """ True airspeed to equivalent airspeed conversion for numpy arrays.

        Arguments:
        - tas: True airspeed [m/s]
        - h: Altitude [m]

        Returns:
        - eas: Equivalent airspeed [m/s]
    """
    rho = vdensity(h)
    eas = tas * np.sqrt(rho / rho0)
    return eas


def vcas2tas(cas, h):
    """ Calibrated to true airspeed conversion for numpy arrays.

        Arguments:
        - cas: Calibrated airspeed [m/s]
        - h: Altitude [m]

        Returns:
        - tas: True airspeed [m/s]
    """
    p, rho, _ = vatmos(h)
    qdyn = p0 * ((1.0 + rho0 * cas * cas / (7.0 * p0)) ** 3.5 - 1.0)
    tas = np.sqrt(7.0 * p / rho * ((1.0 + qdyn / p) ** (2.0 / 7.0) - 1.0))

    # cope with negative speed
    tas = np.where(cas < 0, -1 * tas, tas)
    return tas


def vtas2cas(tas, h):
    """ True to calibrated airspeed conversion for numpy arrays.

        Arguments:
        - tas: True airspeed [m/s]
        - h: Altitude [m]

        Returns:
        cas: Calibrated airspeed [m/s]
    """
    p, rho, _ = vatmos(h)
    qdyn = p*((1.+rho*tas*tas/(7.*p))**3.5-1.)
    cas = np.sqrt(7.*p0/rho0*((qdyn/p0+1.)**(2./7.)-1.))

    # cope with negative speed
    cas = np.where(tas<0, -1*cas, cas)
    return cas


def vmach2cas(mach, h):
    """ Mach to calibrated airspeed conversion for numpy arrays.

        Arguments:
        - mach: Mach number [-]
        - h: Altitude [m]

        Returns:
        - cas: Calibrated airspeed [m/s]
    """
    tas = vmach2tas(mach, h)
    cas = vtas2cas(tas, h)
    return cas


def vcas2mach(cas, h):
    """ Calibrated airspeed to Mach conversion for numpy arrays.

        Arguments:
        - cas: Calibrated airspeed [m/s]
        - h: Altitude [m]

        Returns:
        - mach: Mach number [-]
    """
    tas = vcas2tas(cas, h)
    M   = vtas2mach(tas, h)
    return M

def vcasormach(spd, h):
    """ Interpret input speed as either CAS or a Mach number, and return TAS, CAS, and Mach.

        Arguments:
        - spd: Airspeed. Interpreted as Mach number [-] when its value is below the
               CAS/Mach threshold. Otherwise interpreted as CAS [m/s].
        - h: Altitude [m]

        Returns:
        - tas: True airspeed [m/s]
        - cas: Calibrated airspeed [m/s]
        - mach: Mach number [-]
    """
    ismach = np.logical_and(spd > 0.1, spd < casmach_thr)
    tas = np.where(ismach, vmach2tas(spd, h), vcas2tas(spd, h))
    cas = np.where(ismach, vtas2cas(tas, h), spd)
    mach   = np.where(ismach, spd, vtas2mach(tas, h))
    return tas, cas, mach


def vcasormach2tas(spd, h):
    """ Interpret input speed as either CAS or a Mach number, and return TAS.

        Arguments:
        - spd: Airspeed. Interpreted as Mach number [-] when its value is below the
               CAS/Mach threshold. Otherwise interpreted as CAS [m/s].
        - h: Altitude [m]

        Returns:
        - tas: True airspeed [m/s]
    """
    ismach = np.logical_and(spd > 0.1, spd < casmach_thr)
    return np.where(ismach, vmach2tas(spd, h), vcas2tas(spd, h))


def crossoveralt(cas, mach):
    """ Calculate crossover altitude for given CAS and Mach number.

        Calculates the altitude where the given CAS and Mach values
        correspond to the same true airspeed.

        (BADA User Manual 3.12, p. 12)

        Arguments:
        - cas: Calibrated airspeed [m/s]
        - mach: Mach number [-]

        Returns:
        - Altitude [m].
    """
    # Delta: pressure ratio at the transition altitude
    delta = (((1.0 + 0.5 * (gamma - 1.0) * (cas / a0) ** 2) **
                (gamma / (gamma - 1.0)) - 1.0) /
                ((1.0 + 0.5 * (gamma - 1.0) * mach ** 2) **
                (gamma / (gamma - 1.0)) - 1.0))
    # Theta: Temperature ratio at the transition altitude
    theta = delta ** (-beta * R / g0)
    return 1000.0 / 6.5 * T0 * (1.0 - theta)

# ------------------------------------------------------------------------------
# Scalar aero functions
# ------------------------------------------------------------------------------
def atmos(h):
    """ atmos(altitude): International Standard Atmosphere calculator

        Input:
              h =  altitude in meters 0.0 < h < 84852.
        (will be clipped when outside range, integer input allowed)
        Output:
              [p,rho,T]    (in SI-units: Pa, kg/m3 and K) """

    # Constants

    # Base values and gradient in table from hand-out
    # (but corrected to avoid small discontinuities at borders of layers)
    h0 = [0.0, 11000., 20000., 32000., 47000., 51000., 71000., 86852.]

    p0 = [101325.,                  # Sea level
           22631.7009099,           # 11 km
            5474.71768857,          # 20 km
             867.974468302,         # 32 km
             110.898214043,         # 47 km
              66.939,               # 51 km
               3.9564 ]             # 71 km

    T0 = [288.15,  # Sea level
          216.65,  # 11 km
          216.65,  # 20 km
          228.65,  # 32 km
          270.65,  # 47 km
          270.65,  # 51 km
          214.65]  # 71 km

    # a = lapse rate (temp gradient)
    # integer 0 indicates isothermic layer!
    a  = [-0.0065, # 0-11 km
            0,     # 11-20 km
          0.001,   # 20-32 km
          0.0028,  # 32-47 km
            0,     # 47-51 km
          -0.0028, # 51-71 km
          -0.002]  # 71-   km

    # Clip altitude to maximum!
    h = max(0.0, min(float(h), h0[-1]))

    # Find correct layer
    i = 0
    while h > h0[i+1] and i < len(h0) - 2:
        i = i+1

    # Calculate if sothermic layer
    if a[i] == 0:
        T   = T0[i]
        p   = p0[i]*exp(-g0/(R*T)*(h-h0[i]))
        rho = p/(R*T)

    # Calculate for temperature gradient
    else:
        T   = T0[i] + a[i]*(h-h0[i])
        p   = p0[i]*((T/T0[i])**(-g0/(a[i]*R)))
        rho = p/(R*T)

    return p, rho, T


def temp(h):
    """ temp (altitude): Temperature only version of ISA atmos

        Input:
              h =  altitude in meters 0.0 < h < 84852.
        (will be clipped when outside range, integer input allowed)
        Output:
              T    (in SI-unit: K """

    # Base values and gradient in table from hand-out
    # (but corrected to avoid small discontinuities at borders of layers)
    h0 = [0.0, 11000., 20000., 32000., 47000., 51000., 71000., 86852.]

    T0 = [288.15,  # Sea level
          216.65,  # 11 km
          216.65,  # 20 km
          228.65,  # 32 km
          270.65,  # 47 km
          270.65,  # 51 km
          214.65]  # 71 km

    # a = lapse rate (temp gradient)
    # integer 0 indicates isothermic layer!
    a  = [-0.0065, # 0-11 km
            0 ,    # 11-20 km
          0.001,   # 20-32 km
          0.0028,  # 32-47 km
            0 ,    # 47-51 km
          -0.0028, # 51-71 km
          -0.002]  # 71-   km

    # Clip altitude to maximum!
    h = max(0.0,min(float(h),h0[-1]))


    # Find correct layer
    i = 0
    while h>h0[i+1] and i<len(h0)-2:
        i = i+1

    # Calculate if sothermic layer
    if a[i]==0:
        T   = T0[i]

    # Calculate for temperature gradient
    else:
        T   = T0[i] + a[i]*(h-h0[i])

    return T


# Atmos wrappings:
def pressure(h):          # h [m]
    p, r, T = atmos(h)
    return p


def density(h):   # air density at given altitude h [m]
    p, r, T = atmos(h)
    return r


def vsound(h):  # Speed of sound for given altitude h [m]
    T = temp(h)
    a = sqrt(gamma*R*T)
    return a


# ---------Speed conversions---h in [m]------------------
def tas2mach(tas, h):
    """ True airspeed (tas) to mach number conversion """
    a = vsound(h)
    M = tas / a
    return M


def mach2tas(M, h):
    """ True airspeed (tas) to mach number conversion """
    a = vsound(h)
    tas = M * a
    return tas


def eas2tas(eas, h):
    """ Equivalent airspeed to true airspeed """
    rho = density(h)
    tas = eas * sqrt(rho0 / rho)
    return tas


def tas2eas(tas, h):
    """ True airspeed to equivent airspeed """
    rho = density(h)
    eas = tas * sqrt(rho / rho0)
    return eas


def cas2tas(cas, h):
    """ cas2tas conversion both m/s h in m """
    p, rho, T = atmos(h)
    qdyn = p0*((1.+rho0*cas*cas/(7.*p0))**3.5-1.)
    tas = sqrt(7.*p/rho*((1.+qdyn/p)**(2./7.)-1.))
    tas = -1 * tas if cas < 0 else tas
    return tas


def tas2cas(tas, h):
    """ tas2cas conversion both m/s """
    p, rho, T = atmos(h)
    qdyn = p*((1.+rho*tas*tas/(7.*p))**3.5-1.)
    cas = sqrt(7.*p0/rho0*((qdyn/p0+1.)**(2./7.)-1.))
    cas = -1 * cas if tas < 0 else cas
    return cas


def mach2cas(M, h):
    """ Mach to CAS conversion """
    tas = mach2tas(M, h)
    cas = tas2cas(tas, h)
    return cas


def cas2mach(cas, h):
    """ CAS Mach conversion """
    tas = cas2tas(cas, h)
    M   = tas2mach(tas, h)
    return M

def casormach(spd,h):
    if 0.1 < spd < casmach_thr:
        # Interpret spd as Mach number
        tas = mach2tas(spd, h)
        cas = mach2cas(spd, h)
        m   = spd
    else:
        # Interpret spd as CAS
        tas = cas2tas(spd,h)
        cas = spd
        m   = cas2mach(spd, h)
    return tas, cas, m

def casormach2tas(spd,h):
    if 0.1 < spd < casmach_thr:
        # Interpret spd as Mach number
        tas = mach2tas(spd, h)
    else:
        # Interpret spd as CAS
        tas = cas2tas(spd,h)
    return tas


def metres_to_feet_rounded(metres):
    """
    Converts metres to feet.
    Returns feet as rounded integer.
    """
    return int(round(metres / ft))


def metric_spd_to_knots_rounded(speed):
    """
    Converts speed in m/s to knots.
    Returns knots as rounded integer.
    """
    return int(round(speed / kts))



""" Polygon functions """
import numpy as np
import OpenGL.GLU as glu


class PolygonSet:
    """ Generate one or more polygons from their contours. """
    in_poly = False
    in_contour = False
    tess = None

    def __init__(self):
        self.vbuf = []
        self.prevnode = None
        self.prevcp = None
        self.start_vertex = None
        self.start_cp = None

        if not PolygonSet.tess:
            PolygonSet.tess = glu.gluNewTess()
            glu.gluTessCallback(PolygonSet.tess, glu.GLU_TESS_VERTEX_DATA, lambda vertex, vbuf: vbuf.extend(vertex[0:2]))
            glu.gluTessCallback(PolygonSet.tess, glu.GLU_EDGE_FLAG, lambda flag: None)
            glu.gluTessCallback(PolygonSet.tess, glu.GLU_TESS_COMBINE, lambda c, d, w: np.array(c))

    def bufsize(self):
        return len(self.vbuf)

    def begin(self):
        self.end()
        glu.gluTessBeginPolygon(PolygonSet.tess, self.vbuf)
        PolygonSet.in_poly = True

    def end(self):
        if PolygonSet.in_poly:
            self.endContour()
            glu.gluEndPolygon(PolygonSet.tess)
            PolygonSet.in_poly = False

    def beginContour(self):
        self.endContour()
        if not PolygonSet.in_poly:
            self.begin()
        glu.gluTessBeginContour(PolygonSet.tess)
        PolygonSet.in_contour = True

    def endContour(self):
        if PolygonSet.in_contour:
            if self.prevcp is not None or self.start_cp is not None:
                self.addVertex(self.start_vertex, self.start_cp)
            glu.gluTessEndContour(PolygonSet.tess)
            PolygonSet.in_contour = False
            self.prevcp = None
            self.prevnode = None
            self.start_vertex = None
            self.start_cp = None

    def addContour(self, contour):
        self.begin()
        for vertex in np.reshape(contour, (len(contour) // 2, 2)):
            self.addVertex(np.append(vertex, 0.0))
        self.end()

    def addVertex(self, vertex, controlpoint=None):
        if not PolygonSet.in_contour:
            self.beginContour()
            self.start_vertex = vertex
            self.start_cp = controlpoint
            glu.gluTessVertex(PolygonSet.tess, vertex, vertex)

        elif abs(vertex[0] - self.prevnode[0]) >= 1e-7 or abs(vertex[1] - self.prevnode[1]) >= 1e-7:
            if (controlpoint is None and self.prevcp is None):
                glu.gluTessVertex(PolygonSet.tess, vertex, vertex)
            else:
                if controlpoint is not None:
                    if self.prevcp is not None:
                        self.bezier2(vertex, 2 * vertex - controlpoint)
                    else:
                        self.bezier1(vertex, 2 * vertex - controlpoint)

                else:
                    self.bezier1(vertex, self.prevcp)

        self.prevnode = vertex
        self.prevcp = controlpoint

    def bezier1(self, vertex, controlpoint):
        for fraction in [0.2, 0.4, 0.6, 0.8, 1.0]:
            lnode1 = self.prevnode + fraction * (controlpoint - self.prevnode)
            lnode2 = controlpoint + fraction * (vertex - controlpoint)

            vnew = lnode1 + fraction * (lnode2 - lnode1)
            glu.gluTessVertex(PolygonSet.tess, vnew, vnew)

    def bezier2(self, vertex, controlpoint):
        for fraction in [0.2, 0.4, 0.6, 0.8, 1.0]:
            lnode1 = self.prevnode + fraction * (self.prevcp - self.prevnode)
            lnode2 = self.prevcp + fraction * (controlpoint - self.prevcp)
            lnode3 = controlpoint + fraction * (vertex - controlpoint)

            lnode4 = lnode1 + fraction * (lnode2 - lnode1)
            lnode5 = lnode2 + fraction * (lnode3 - lnode2)

            vnew = lnode4 + fraction * (lnode5 - lnode4)
            glu.gluTessVertex(PolygonSet.tess, vnew, vnew)


class BoundingBox:
    """ Calculate bounding box for a set of vertices. """

    def __init__(self):
        self.corners = [999.9, -999.9, 999.9, -999.9]

    def update(self, vertex):
        self.corners[0] = min(self.corners[0], vertex[0])
        self.corners[1] = max(self.corners[1], vertex[0])
        self.corners[2] = min(self.corners[2], vertex[1])
        self.corners[3] = max(self.corners[3], vertex[1])

    def center(self):
        return [0.5 * (self.corners[0] + self.corners[1]),
                0.5 * (self.corners[2] + self.corners[3])]



import numpy as np

from bluesky.core import Base
from bluesky.network.sharedstate import ActData


class TrafficProxy(Base):
    ''' Proxy object implementation for Traffic()-like access
        of received data on client side.
    '''
    id: ActData[list] = ActData(group='acdata')
    lat: ActData[np.ndarray] = ActData(0, group='acdata')
    lon: ActData[np.ndarray] = ActData(0, group='acdata')
    alt: ActData[np.ndarray] = ActData(0, group='acdata')
    trk: ActData[np.ndarray] = ActData(0, group='acdata')
    cas: ActData[np.ndarray] = ActData(0, group='acdata')
    tas: ActData[np.ndarray] = ActData(0, group='acdata')
    gs: ActData[np.ndarray] = ActData(0, group='acdata')
    vs: ActData[np.ndarray] = ActData(0, group='acdata')
    translvl: ActData[float] = ActData(0.0, group='acdata')

    # data['ingroup']    = bs.traf.groups.ingroup
    # data['inconf'] = bs.traf.cd.inconf
    # data['tcpamax'] = bs.traf.cd.tcpamax
    # data['rpz'] = bs.traf.cd.rpz
    # data['nconf_cur'] = len(bs.traf.cd.confpairs_unique)
    # data['nconf_tot'] = len(bs.traf.cd.confpairs_all)
    # data['nlos_cur'] = len(bs.traf.cd.lospairs_unique)
    # data['nlos_tot'] = len(bs.traf.cd.lospairs_all)
    
    # data['vmin']       = bs.traf.perf.vmin
    # data['vmax']       = bs.traf.perf.vmax


    # # Send casmachthr for route visualization
    # data['casmachthr']    = aero.casmach_thr

    # ASAS resolutions for visualization. Only send when evaluated
    # data['asastas']  = bs.traf.cr.tas
    # data['asastrk']  = bs.traf.cr.trk

    # # Aircraft (group) color
    # if self.custacclr:
    #     data['custacclr'] = self.custacclr
    # if self.custgrclr:
    #     data['custgrclr'] = self.custgrclr

    @property
    def ntraf(self):
        return len(self.id)

    def id2idx(self, acid):
        """Find index of aircraft id"""
        if not isinstance(acid, str):
            # id2idx is called for multiple id's
            # Fast way of finding indices of all ACID's in a given list
            tmp = dict((v, i) for i, v in enumerate(self.id))
            return [tmp.get(acidi, -1) for acidi in acid]
        else:
             # Catch last created id (* or # symbol)
            if acid in ('#', '*'):
                return self.ntraf - 1

            try:
                return self.id.index(acid.upper())
            except:
                return -1



''' Very incomplete implementation of a DirectDraw texture file loader.
    Basically only able to load BlueSky's world texture, which is in DXT1 format.

    DDS file format taken from
    https://docs.microsoft.com/en-us/windows/desktop/direct3ddds/dx-graphics-dds-pguide
'''
from ctypes import c_char, c_uint32, Structure, sizeof

# DDS flags
DDSD_CAPS = 0x1
DDSD_HEIGHT = 0x2
DDSD_WIDTH = 0x4
DDSD_PITCH = 0x8
DDSD_PIXELFORMAT = 0x1000
DDSD_MIPMAPCOUNT = 0x20000
DDSD_LINEARSIZE = 0x80000
DDSD_DEPTH = 0x800000

# Pixel format flags
DDPF_ALPHAPIXELS = 0x1
DDPF_ALPHA = 0x2
DDPF_FOURCC = 0x4
DDPF_RGB = 0x40
DDPF_YUV = 0x200
DDPF_LUMINANCE = 0x20000

class PixelFormat(Structure):
    _fields_ = [('dwSize', c_uint32), ('dwFlags', c_uint32),
        ('dwFourCC', c_char * 4), ('dwRGBBitCount', c_uint32),
        ('dwRBitMask', c_uint32), ('dwGBitMask', c_uint32),
        ('dwBBitMask', c_uint32), ('dwABitMask', c_uint32)]


class DDSHeader(Structure):
    _fields_ = [('dwSize', c_uint32), ('dwFlags', c_uint32),
        ('dwHeight', c_uint32), ('dwWidth', c_uint32),
        ('dwPitchOrLinearSize', c_uint32), ('dwDepth', c_uint32),
        ('dwMipMapCount', c_uint32), ('dwReserved1', c_uint32 * 11),
        ('ddspf', PixelFormat), ('dwCaps', c_uint32),
        ('dwCaps2', c_uint32), ('dwCaps3', c_uint32),
        ('dwCaps4', c_uint32), ('dwReserved2', c_uint32)]


class DX10Header(Structure):
    _fields_ = [('dxgiFormat', c_uint32), ('resourceDimension', c_uint32),
        ('miscFlag', c_uint32), ('arraySize', c_uint32),
        ('miscFlags2', c_uint32)]


class DDSError(Exception):
    pass


class DDSTexture:
    ''' Loader class for Direct Draw texture files. '''
    def __init__(self, fname):
        header = DDSHeader()
        dx10header = None
        with open(fname, 'rb') as fin:
            # Read magic number
            magic = fin.read(4)
            if magic != b'DDS ':
                raise DDSError('File not recognised as DDS texure')

            # Read DDS file header
            size = fin.readinto(header)
            if size != sizeof(header):
                raise DDSError('File not recognised as DDS texure')

            # If extended format, extract extended header
            if header.ddspf.dwFourCC == b'DX10':
                dx10header = DX10Header()
                size = fin.readinto(dx10header)
                if size != sizeof(dx10header):
                    raise DDSError('File not recognised as DDS texure')

            # Keep useful attributes
            self.width = header.dwWidth
            self.height = header.dwHeight
            if header.dwCaps & DDSD_MIPMAPCOUNT:
                self.mipmapcount = header.dwMipMapCount
            else:
                self.mipmapcount = 0

            # load data
            self.is_compressed = (header.ddspf.dwFlags & DDPF_FOURCC != 0)

            # Calculate size of current texture level
            calcsize = lambda w, h: max(1, ((w + 3) // 4)) * max(1, ((h + 3) // 4)) * 8

            # Read main texture and possible mipmaps
            w, h = self.width, self.height
            self.data = fin.read(calcsize(w, h))
            self.mipmaps = list()
            for _ in range(self.mipmapcount):
                w = w // 2
                h = h // 2
                self.mipmaps.append((tsize, fin.read(calcsize(w, h))))



import pygame as pg
from .fastfont import Fastfont

black = (0, 0, 0)
white    = (255,255,255)
darkgrey = (25, 25, 48)
grey = (84, 84, 114)
darkblue = (25, 25, 64, 100)
white = (255, 255, 255)
green = (0, 255, 0)
blue = (0, 0, 255)
red = (255, 0, 0)
lightgreyblue = (130, 150, 190)  # waypoint symbol color
lightgreygreen = (149, 215, 179)  # grid color
lightcyan = (0, 255, 255)  # FIR boundaries
amber = (255,163,71)  # Conflicting aircraft
magenta = (255,0,255) # Used for route

class Console:
    """ 
    Console (aka EditWin) class definition : Edit window & console class

    Methods:
        echo(msg)         : Print a message
        insert(message)   : insert characters in current edit line
        backspace()       : process backspace
        getline()         : return current edit line
        enter()           : enter, end of edit line
        scroll()          : scroll up one line
        update()          : redraw update bitmap of edit window

    Created by  : Jacco M. Hoekstra (TU Delft)
    """

    def __init__(self,win,nch,nlin,winx,winy):
        # Was Helvetica,14
        self.fontedit = Fastfont(win,'Courier New',14,white,False,False) # name, size, bold,italic

        # Edit window: 6 line of 64 chars
        self.content = []
        self.nch = nch    # number of chars per line
        self.nlin = nlin   # number of lines in windows
        self.winx = winx   # x-coordinate in pixels of left side
        self.winy = winy - self.nlin*self.fontedit.linedy   # y-coordinate in pixels of top
        self.msg = []        # Messages in edit window
        
        for i in range(self.nlin):
            line= self.nch*[' ']
            self.content.append(line)
        self.content0 = self.content
        
        self.xcursor = 0
        self.xedit = 0
        # self.printeditwin('Testing 1,2,3')
        self.bmpdy = self.nlin*self.fontedit.linedy
        self.bmpdx = int(self.nch*self.fontedit.linedy*10/17) + 2 # Guess max aspect ratio
        self.bmp = pg.Surface([self.bmpdx,self.bmpdy],
                                   pg.SRCALPHA, 32)
                                   
        self.bmp.fill(darkblue)
        self.rect = pg.Rect(self.winx,self.winy,
                             self.bmpdx,self.bmpdy)
        self.redraw = True
     
        return


    def echo(self,msg): 
        """print a message to console window"""
        if self.xedit==self.xcursor:
            self.insert(msg)
            j = int(self.xcursor/self.nch)      
            self.xcursor = (j+1)*self.nch
            self.xedit = self.xcursor
    
            # Check for End of window
            if self.xedit >= (self.nch-1)*(self.nlin-1):
                del self.content[0]
                self.content.append(self.nch*[' '])
                self.xcursor = j*self.nch
                self.xedit = self.xcursor
        else:
            self.msg.append(msg)   # buffer
        return


    def insert(self,message):
        i = self.xcursor%self.nch
        j = int(self.xcursor/self.nch)
        for ich in range(len(message)):
            self.content[j][i]=message[ich]
            i = i+1
            # Check for end-of line
            if i>=self.nch:
                i = 0
                j = j+1
                # Check for end-of edit window
                if j>=self.nlin:
                    self.scroll()
                    j = j-1
        self.xcursor = j*self.nch+i
        self.redraw = True
        return

    def backspace(self):
        if self.xcursor>self.xedit:
            self.xcursor = self.xcursor-1
        self.redraw = True
        i = self.xcursor%self.nch
        j = int(self.xcursor/self.nch)
        self.content[j][i]=" "
        return

    def getline(self): # enter was pressed ro we need current command line
        line = ""
        for idx in range(self.xedit,self.xcursor):        
            i = idx%self.nch
            j = int(idx/self.nch)
            line = line+self.content[j][i]
        return line

    def enter(self):           
        j = int(self.xcursor/self.nch)        
        self.xcursor = (j+1)*self.nch
        self.xedit = self.xcursor

        # End of window
        if self.xedit >= (self.nch-1)*(self.nlin-1):
            del self.content[0]
            self.content.append(self.nch*[' '])
            self.xcursor = j*self.nch
            self.xedit = self.xcursor

        # Print buffered messages
        self.redraw = True
        while len(self.msg)>0:
            self.echo(self.msg[0])  # No endless recursion becasue xedit==xcursor
            del self.msg[0]
        return

    def scroll(self):
        """Scroll window"""
        del self.content[0]
        self.content.append(self.nch*[' '])
        self.xcursor = self.xcursor-self.nch
        self.xedit = self.xedit-self.nch

    def update(self):
        """Update: Draw a new frame"""
        # Draw edit window
        if self.redraw:
            self.bmp.fill(darkgrey)
            
            for j in range(self.nlin):
                for i in range(self.nch):
                    if True or self.content[j][i] != self.content0[j][i]:
                        x = i*int(self.fontedit.linedy*10/17) + 1
                        y = j*self.fontedit.linedy+int(self.fontedit.linedy/6)
                        self.fontedit.printat(self.bmp,
                                              x,y,
                                              self.content[j][i])
                        self.content0[j][i]=self.content[j][i]
            # Draw cursor
            i = self.xcursor%self.nch
            j = int(self.xcursor/self.nch)
            x = i*int(self.fontedit.linedy*10/17)
            y = j*self.fontedit.linedy+int(self.fontedit.linedy/6)
            self.fontedit.printat(self.bmp,x,y,"_")
            self.bmp.set_alpha(127)    
            self.redraw = False

        return



''' Traffic OpenGL visualisation. '''
import numpy as np
from bluesky.ui.qtgl import glhelpers as glh

from bluesky.stack import command
from bluesky.tools import geo
from bluesky import settings
from bluesky.ui import palette
from bluesky.tools.aero import ft, nm, kts
from bluesky.network import context as ctx
from bluesky.network.subscriber import subscriber
from bluesky.network.sharedstate import ActData


# Register settings defaults
settings.set_variable_defaults(
    text_size=13, ac_size=16,
    asas_vmin=200.0, asas_vmax=500.0)

palette.set_default_colours(
    aircraft=(0, 255, 0),
    conflict=(255, 160, 0),
    route=(255, 0, 255),
    trails=(0, 255, 255))

# Static defines
MAX_NAIRCRAFT = 10000
MAX_NCONFLICTS = 25000
MAX_ROUTE_LENGTH = 500
ROUTE_SIZE = 500
TRAILS_SIZE = 1000000


class Traffic(glh.RenderObject, layer=100):
    ''' Traffic OpenGL object. '''
    # Per remote node attributes
    show_pz: ActData[bool] = ActData(False)
    show_traf: ActData[bool] = ActData(True)
    show_lbl: ActData[int] = ActData(2)
    ssd_all: ActData[bool] = ActData(False)
    ssd_conflicts: ActData[bool] = ActData(False)
    ssd_ownship: ActData[set] = ActData(set())
    altrange: ActData[tuple] = ActData(tuple())
    naircraft: ActData[int] = ActData(0)
    zoom: ActData[float] = ActData(1.0, group='panzoom')

    @command
    def showpz(self, flag:bool=None):
        ''' Toggle drawing of aircraft protected zones. '''
        # TODO: add to SWRAD (flag=SYM)
        self.show_pz = not self.show_pz if flag is None else flag

    @command
    def showtraf(self, flag:bool=None):
        ''' Toggle drawing of aircraft. '''
        self.show_traf = not self.show_traf if flag is None else flag

    @command
    def label(self, flag:int=None):
        ''' Toggle drawing of aircraft label. '''
        # Cycle aircraft label through detail level 0,1,2
        if flag is None:
            self.show_lbl = (self.show_lbl + 1) % 3

        # Or use the argument if it is an integer
        else:
            self.show_lbl = min(2,max(0,flag))

    @command(name='SSD')
    def showssd(self, *arg:'txt'):
        ''' Show/hide SSD (state-space diagram) for one or more, or all aircraft.
        
            The SSD shows the vector space of conflicting vs. conflict-free aircraft
            velocities.

            Arguments:
            ALL: Show SSD for all aircraft
            CONFLICTS: Show SSD only for aircraft in conflict
            OFF: Hide SSDs
            acid(s): Show/hide SSD for given aircraft
        '''
        if 'ALL' in arg:
            self.ssd_all      = True
            self.ssd_conflicts = False
        elif 'CONFLICTS' in arg:
            self.ssd_all      = False
            self.ssd_conflicts = True
        elif 'OFF' in arg:
            self.ssd_all      = False
            self.ssd_conflicts = False
            self.ssd_ownship = set()
        else:
            remove = self.ssd_ownship.intersection(arg)
            self.ssd_ownship = self.ssd_ownship.union(arg) - remove

    @command
    def filteralt(self, flag:bool=None, bottom:'alt'=-1e99, top:'alt'=1e99):
        ''' Display aircraft on only a selected range of altitudes

            Arguments:
            - flag: Turn altitude filtering ON/OFF
            - bottom: The lowest altitude in the visible range (optional)
            - top: The highest altitude in the visible range (optional)
        '''
        if flag is None:
            if not self.altrange:
                return True, f'The current altitude range is unlimited'
            elif self.altrange[0] < -1e90:
                return True, f'The current altitude range is limited below {self.altrange[1]} meters'
            elif self.altrange[1] > 1e90:
                return True, f'The current altitude range is limited above {self.altrange[0]} meters'
            else:
                return True, f'The current altitude range is limited between {self.altrange[0]} and {self.altrange[1]} meters'
        if flag:
            self.altrange = (bottom, top)        

    def __init__(self, parent=None):
        super().__init__(parent)
        self.initialized = False
        self.route_acid = ''
        self.asas_vmin = settings.asas_vmin
        self.asas_vmax = settings.asas_vmax
        self.hdg = glh.GLBuffer()
        self.rpz = glh.GLBuffer()
        self.lat = glh.GLBuffer()
        self.lon = glh.GLBuffer()
        self.alt = glh.GLBuffer()
        self.tas = glh.GLBuffer()
        self.color = glh.GLBuffer()
        self.lbl = glh.GLBuffer()
        self.asasn = glh.GLBuffer()
        self.asase = glh.GLBuffer()

        self.ssd = glh.VertexArrayObject(glh.gl.GL_POINTS, shader_type='ssd')
        self.protectedzone = glh.Circle()
        self.ac_symbol = glh.VertexArrayObject(glh.gl.GL_TRIANGLE_FAN)
        self.aclabels = glh.Text(settings.text_size, (8, 3))
        self.cpalines = glh.VertexArrayObject(glh.gl.GL_LINES)
        self.route = glh.VertexArrayObject(glh.gl.GL_LINES)
        self.routelbl = glh.Text(settings.text_size, (12, 2))
        self.rwaypoints = glh.VertexArrayObject(glh.gl.GL_LINE_LOOP)
        self.traillines = glh.VertexArrayObject(glh.gl.GL_LINES)

    def create(self):
        ac_size = settings.ac_size
        wpt_size = settings.wpt_size
        self.hdg.create(MAX_NAIRCRAFT * 4, glh.GLBuffer.UsagePattern.StreamDraw)
        self.lat.create(MAX_NAIRCRAFT * 4, glh.GLBuffer.UsagePattern.StreamDraw)
        self.lon.create(MAX_NAIRCRAFT * 4, glh.GLBuffer.UsagePattern.StreamDraw)
        self.alt.create(MAX_NAIRCRAFT * 4, glh.GLBuffer.UsagePattern.StreamDraw)
        self.tas.create(MAX_NAIRCRAFT * 4, glh.GLBuffer.UsagePattern.StreamDraw)
        self.color.create(MAX_NAIRCRAFT * 4, glh.GLBuffer.UsagePattern.StreamDraw)
        self.lbl.create(MAX_NAIRCRAFT * 24, glh.GLBuffer.UsagePattern.StreamDraw)
        self.asasn.create(MAX_NAIRCRAFT * 24, glh.GLBuffer.UsagePattern.StreamDraw)
        self.asase.create(MAX_NAIRCRAFT * 24, glh.GLBuffer.UsagePattern.StreamDraw)
        self.rpz.create(MAX_NAIRCRAFT * 4, glh.GLBuffer.UsagePattern.StreamDraw)

        self.ssd.create(lat1=self.lat, lon1=self.lon, alt1=self.alt,
                        tas1=self.tas, trk1=self.hdg)
        self.ssd.set_attribs(selssd=MAX_NAIRCRAFT, instance_divisor=1,
                             datatype=glh.gl.GL_UNSIGNED_BYTE, normalize=True)
        self.ssd.set_attribs(lat0=self.lat, lon0=self.lon,
                             alt0=self.alt, tas0=self.tas,
                             trk0=self.hdg, asasn=self.asasn,
                             asase=self.asase, instance_divisor=1)

        #self.protectedzone.create(radius=1.0)
        self.protectedzone.create(radius=0.5)
        self.protectedzone.set_attribs(lat=self.lat, lon=self.lon, scale=self.rpz,
                                       color=self.color, instance_divisor=1)

        acvertices = np.array([(0.0, 0.5 * ac_size), (-0.5 * ac_size, -0.5 * ac_size),
                               (0.0, -0.25 * ac_size), (0.5 * ac_size, -0.5 * ac_size)],
                              dtype=np.float32)
        self.ac_symbol.create(vertex=acvertices)

        self.ac_symbol.set_attribs(lat=self.lat, lon=self.lon, color=self.color,
                                   orientation=self.hdg, instance_divisor=1)

        self.aclabels.create(self.lbl, self.lat, self.lon, self.color,
                             (ac_size, -0.5 * ac_size), instanced=True)

        self.cpalines.create(vertex=MAX_NCONFLICTS * 16, color=palette.conflict, usage=glh.GLBuffer.UsagePattern.StreamDraw)

        # ------- Aircraft Route -------------------------
        self.route.create(vertex=ROUTE_SIZE * 8, color=palette.route, usage=glh.GLBuffer.UsagePattern.DynamicDraw)

        self.routelbl.create(ROUTE_SIZE * 24, ROUTE_SIZE * 4, ROUTE_SIZE * 4,
                             palette.route, (wpt_size, 0.5 * wpt_size), instanced=True)
        rwptvertices = np.array([(-0.2 * wpt_size, -0.2 * wpt_size),
                                 (0.0,            -0.8 * wpt_size),
                                 (0.2 * wpt_size, -0.2 * wpt_size),
                                 (0.8 * wpt_size,  0.0),
                                 (0.2 * wpt_size,  0.2 * wpt_size),
                                 (0.0,             0.8 * wpt_size),
                                 (-0.2 * wpt_size,  0.2 * wpt_size),
                                 (-0.8 * wpt_size,  0.0)], dtype=np.float32)
        self.rwaypoints.create(vertex=rwptvertices, color=palette.route)
        self.rwaypoints.set_attribs(lat=self.routelbl.lat, lon=self.routelbl.lon, instance_divisor=1)

        # # --------Aircraft Trails------------------------------------------------
        self.traillines.create(vertex=TRAILS_SIZE * 16, color=palette.trails)
        self.initialized = True

    def draw(self):
        ''' Draw all traffic graphics. '''
        # Get data for active node
        if self.naircraft == 0 or not self.show_traf:
            return

        # Send the (possibly) updated global uniforms to the buffer
        self.shaderset.set_vertex_scale_type(self.shaderset.VERTEX_IS_LATLON)
        self.shaderset.enable_wrap(False)

        self.route.draw()
        self.cpalines.draw()
        self.traillines.draw()

        # --- DRAW THE INSTANCED AIRCRAFT SHAPES ------------------------------
        # update wrap longitude and direction for the instanced objects
        self.shaderset.enable_wrap(True)

        # PZ circles only when they are bigger than the A/C symbols
        if self.show_pz and self.zoom >= 0.15:
            self.shaderset.set_vertex_scale_type(
                self.shaderset.VERTEX_IS_METERS)
            self.protectedzone.draw(n_instances=self.naircraft)

        self.shaderset.set_vertex_scale_type(self.shaderset.VERTEX_IS_SCREEN)

        # Draw traffic symbols
        self.ac_symbol.draw(n_instances=self.naircraft)

        if self.routelbl.n_instances:
            self.rwaypoints.draw(n_instances=self.routelbl.n_instances)
            self.routelbl.draw()

        if self.show_lbl:
            self.aclabels.draw(n_instances=self.naircraft)

        # SSD
        if self.ssd_all or self.ssd_conflicts or len(self.ssd_ownship) > 0:
            ssd_shader = glh.ShaderSet.get_shader('ssd')
            ssd_shader.bind()
            glh.gl.glUniform3f(ssd_shader.uniforms['Vlimits'].loc, self.asas_vmin **
                           2, self.asas_vmax ** 2, self.asas_vmax)
            glh.gl.glUniform1i(ssd_shader.uniforms['n_ac'].loc, self.naircraft)
            self.ssd.draw(vertex_count=self.naircraft,
                          n_instances=self.naircraft)

    @subscriber(topic='TRAILS')
    def update_trails_data(self, data):
        ''' Update GPU buffers with route data from simulation. '''
        if not self.initialized:
            return
        if ctx.action == ctx.action.Reset or ctx.action == ctx.action.ActChange:# TODO hack
            # Simulation reset: Clear all entries
            self.traillines.set_vertex_count(0)
            return
        if len(data.traillat0) > 0:
            self.glsurface.makeCurrent()
            self.traillines.set_vertex_count(len(data.traillat0))
            if len(data.traillat0) > 0:
                self.traillines.update(vertex=np.array(
                        list(zip(data.traillat0, data.traillon0,
                                data.traillat1, data.traillon1)), dtype=np.float32))
        else:
            self.traillines.set_vertex_count(0)

    @subscriber(topic='ROUTEDATA', actonly=True)
    def update_route_data(self, data):
        ''' Update GPU buffers with route data from simulation. '''
        if not self.initialized:
            return
        if ctx.action == ctx.action.Reset or ctx.action == ctx.action.ActChange:# TODO hack
            # Simulation reset: Clear all entries
            self.route.set_vertex_count(0)
            self.routelbl.n_instances = 0
            return
        self.glsurface.makeCurrent()
        self.route_acid = data.acid
        if data.acid != "" and len(data.wplat) > 0:
            nsegments = len(data.wplat)
            data.iactwp = min(max(0, data.iactwp), nsegments - 1)
            self.routelbl.n_instances = nsegments
            self.route.set_vertex_count(2 * nsegments)
            routedata = np.empty(4 * nsegments, dtype=np.float32)
            routedata[0:4] = [data.aclat, data.aclon,
                              data.wplat[data.iactwp], data.wplon[data.iactwp]]

            routedata[4::4] = data.wplat[:-1]
            routedata[5::4] = data.wplon[:-1]
            routedata[6::4] = data.wplat[1:]
            routedata[7::4] = data.wplon[1:]

            self.route.update(vertex=routedata)
            wpname = ''
            for wp, alt, spd in zip(data.wpname, data.wpalt, data.wpspd):
                if alt < 0. and spd < 0.:
                    txt = wp[:12].ljust(24)  # No second line
                else:
                    txt = wp[:12].ljust(12)  # Two lines
                    if alt < 0:
                        txt += "-----/"
                    # TODO: get from sim
                    elif alt > 5000.0 * ft:
                        FL = int(round((alt / (100. * ft))))
                        txt += "FL%03d/" % FL
                    else:
                        txt += "%05d/" % int(round(alt / ft))

                    # Speed
                    if spd < 0:
                        txt += "--- "
                    # TODO: get from sim
                    elif spd > settings.casmach_threshold:
                        txt += "%03d" % int(round(spd / kts))
                    else:
                        txt += f"M{spd:.2f}" # Mach number

                wpname += txt.ljust(24)  # Fill out with spaces
            self.routelbl.update(texdepth=np.array(wpname.encode('ascii', 'ignore')),
                                 lat=np.array(data.wplat, dtype=np.float32),
                                 lon=np.array(data.wplon, dtype=np.float32))
        else:
            self.route.set_vertex_count(0)
            self.routelbl.n_instances = 0

    @subscriber(topic='ACDATA', actonly=True)
    def update_aircraft_data(self, data):
        ''' Update GPU buffers with new aircraft simulation data. '''
        if not self.initialized:
            return
        if ctx.action == ctx.action.Reset or ctx.action == ctx.action.ActChange:# TODO hack
            # Simulation reset: Clear all entries
            self.naircraft = 0
            return
        self.glsurface.makeCurrent()
        if self.altrange:
            idx = np.where(
                (data.alt >= self.altrange[0]) * (data.alt <= self.altrange[1]))
            data.lat = data.lat[idx]
            data.lon = data.lon[idx]
            data.trk = data.trk[idx]
            data.alt = data.alt[idx]
            data.tas = data.tas[idx]
            data.vs = data.vs[idx]
            data.rpz = data.rpz[idx]
        naircraft = len(data.lat)
        self.naircraft = naircraft
        # self.asas_vmin = data.vmin # TODO: array should be attribute not uniform
        # self.asas_vmax = data.vmax

        if naircraft == 0:
            self.cpalines.set_vertex_count(0)
        else:
            # Update data in GPU buffers
            self.lat.update(np.array(data.lat, dtype=np.float32))
            self.lon.update(np.array(data.lon, dtype=np.float32))
            self.hdg.update(np.array(data.trk, dtype=np.float32))
            self.alt.update(np.array(data.alt, dtype=np.float32))
            self.tas.update(np.array(data.tas, dtype=np.float32))
            self.rpz.update(np.array(data.rpz, dtype=np.float32))
            if hasattr(data, 'asasn') and hasattr(data, 'asase'):
                self.asasn.update(np.array(data.asasn, dtype=np.float32))
                self.asase.update(np.array(data.asase, dtype=np.float32))

            # CPA lines to indicate conflicts
            ncpalines = np.count_nonzero(data.inconf)

            cpalines = np.zeros(4 * ncpalines, dtype=np.float32)
            self.cpalines.set_vertex_count(2 * ncpalines)

            # Labels and colors
            rawlabel = ''
            color = np.empty(
                (min(naircraft, MAX_NAIRCRAFT), 4), dtype=np.uint8)
            selssd = np.zeros(naircraft, dtype=np.uint8)
            confidx = 0

            custacclr = getattr(data, 'custacclr', dict())
            custgrclr = getattr(data, 'custgrclr', dict())

            zdata = zip(data.id, data.ingroup, data.inconf, data.tcpamax, data.trk, data.gs,
                        data.cas, data.vs, data.alt, data.lat, data.lon)
            for i, (acid, ingroup, inconf, tcpa,
                    trk, gs, cas, vs, alt, lat, lon) in enumerate(zdata):
                if i >= MAX_NAIRCRAFT:
                    break

                # Make label: 3 lines of 8 characters per aircraft
                if self.show_lbl >= 1:
                    rawlabel += '%-8s' % acid[:8]
                    if self.show_lbl == 2:
                        if alt <= data.translvl:
                            rawlabel += '%-5d' % int(alt / ft + 0.5)
                        else:
                            rawlabel += 'FL%03d' % int(alt / ft / 100. + 0.5)
                        vsarrow = 30 if vs > 0.25 else 31 if vs < -0.25 else 32
                        rawlabel += '%1s  %-8d' % (chr(vsarrow),
                                                   int(cas / kts + 0.5))
                    else:
                        rawlabel += 16 * ' '

                if inconf:
                    if self.ssd_conflicts:
                        selssd[i] = 255
                    color[i, :] = palette.conflict + (255,)
                    lat1, lon1 = geo.qdrpos(lat, lon, trk, tcpa * gs / nm)
                    cpalines[4 * confidx: 4 * confidx +
                             4] = [lat, lon, lat1, lon1]
                    confidx += 1
                else:
                    # Get custom color if available, else default
                    rgb = palette.aircraft
                    if ingroup:
                        for groupmask, groupcolor in custgrclr.items():
                            if ingroup & groupmask:
                                rgb = groupcolor
                                break
                    rgb = custacclr.get(acid, rgb)
                    color[i, :] = tuple(rgb) + (255,)

                #  Check if aircraft is selected to show SSD
                if self.ssd_all or acid in self.ssd_ownship:
                    selssd[i] = 255

            if len(self.ssd_ownship) > 0 or self.ssd_conflicts or self.ssd_all:
                self.ssd.update(selssd=selssd)
            self.cpalines.update(vertex=cpalines)
            self.color.update(color)
            self.lbl.update(np.array(rawlabel.encode('utf8'), dtype=np.bytes_))
            
            # If there is a visible route, update the start position
            if self.route_acid in data.id:
                idx = data.id.index(self.route_acid)
                self.route.vertex.update(np.array([data.lat[idx], data.lon[idx]], dtype=np.float32))



""" Wind implementation for BlueSky."""
from numpy import array, sin, cos, arange, radians, ones, append, ndarray, \
                  minimum, repeat, delete, zeros, maximum, floor, interp, \
                  pi, concatenate, unique
from scipy.interpolate import interp1d, RegularGridInterpolator
from bluesky.tools.aero import ft

class Windfield():
    """ Windfield class:
        Methods:
            clear()    = clear windfield, no wind vectors defined

            addpoint(lat,lon,winddir,winddspd,windalt=None)
                       = add a wind vector to a position,
                         windvector can be arrays for altitudes (optional)
                         returns index of vector (0,1,2,3,..)
                         all units are SI units, angles in degrees

            get(lat,lon,alt=0)
                       = get wind vector for given position and optional
                         altitude, all can be arrays,
                         vnorth and veast will be returned in same dimension

            remove(idx) = remove a defined profile using the index

        Members:
            lat(nvec)          = latitudes of wind definitions
            lon(nvec)          = longitudes of wind definitions
            altaxis(nalt)      = altitude axis (fixed, 250 m resolution)

            vnorth(nalt,nvec)  = wind north component [m/s]
            veast(nalt,nvec)   = wind east component [m/s]

            winddim   = Windfield dimension, will automatically be detected:
                          0 = no wind
                          1 = constant wind
                          2 = 2D field (no alt profiles),
                          3 = 3D field (alt dependent wind at some points)

    """
    def __init__(self):
        # For altitude use fixed axis to allow vectorisation later
        self.altmax  = 45000. * ft   # [m]
        self.altstep = 100. * ft    # [m]

        # Axis
        self.altaxis = arange(0., self.altmax + self.altstep, self.altstep)
        self.idxalt  = arange(0, len(self.altaxis), 1.)
        self.nalt    = len(self.altaxis)

        # List of indices of points with an altitude profile (for 3D check)
        self.iprof   = []

        # Clear actual field
        self.clear()
        return

    def clear(self): #Clear actual field
        # Windfield dimension will automatically be detected:
        # 0 = no wind, 1 = constant wind, 2 = 2D field (no alt profiles),
        # 3 = 3D field (alt matters), used to speed up interpolation
        self.winddim = 0
        self.lat     = array([])
        self.lon     = array([])
        self.vnorth  = array([[]])
        self.veast   = array([[]])
        self.nvec    = 0
        self.fe      = None
        self.fn      = None
        return

    def addpointvne(self, lat, lon, vnorth, veast, windalt=None):
        """ Add a vector of lat/lon positions (arrays) with a (2D vector of) 
            wind speed [m/s] in north and east component. 
            Optionally an array with altitudes can be used
        """              
        if windalt is not None and len(windalt) > 1:           
            # Set altitude interpolation functions
            fnorth = interp1d(windalt, vnorth.T, bounds_error=False, 
                              fill_value=(vnorth[0], vnorth[-1]), assume_sorted=True)
            feast  = interp1d(windalt, veast.T, bounds_error=False, 
                              fill_value=(veast[0], veast[-1]), assume_sorted=True)
                       
            # Assume regular grid and set RGI for interpolation
            if len(lat) > 3: 
                try:
                    # Interpolate along windalt axis
                    altaxis = concatenate((array([0.]), windalt))
                    vnaxis = fnorth(altaxis).T
                    veaxis = feast(altaxis).T
                    
                    # Get unique latitudes and longitudes for RGI
                    lats = unique(lat)
                    lons = unique(lon)
                    
                    # Set RGI interpolation functions
                    vevalues = veaxis.reshape((len(altaxis), len(lats), len(lons)))
                    vnvalues = vnaxis.reshape((len(altaxis), len(lats), len(lons)))
                    self.fe = RegularGridInterpolator((altaxis, lats, lons), 
                                                      vevalues, bounds_error=False, fill_value=0.)
                    self.fn = RegularGridInterpolator((altaxis, lats, lons), 
                                                      vnvalues, bounds_error=False, fill_value=0.) 
                except:
                    # Create vn, ve if RGI is not possible
                    vnaxis = fnorth(self.altaxis).T
                    veaxis = feast(self.altaxis).T
            else:
                # Create vn, ve if less than 4 coords are present
                vnaxis = fnorth(self.altaxis).T
                veaxis = feast(self.altaxis).T
        
            self.winddim = 3
            self.iprof.append(len(self.lat) + 1)
        
        else:
            vnaxis = vnorth
            veaxis = veast

        self.nvec += len(lat)
        self.lat = append(self.lat, lat)
        self.lon = append(self.lon, lon)

        if self.vnorth.size == 0:
            self.vnorth = vnaxis
            self.veast  = veaxis
        else:
            self.vnorth = concatenate((self.vnorth, vnaxis), axis=1) 
            self.veast  = concatenate((self.veast, veaxis), axis=1)

        if self.winddim<3: # No 3D => set dim to 0,1 or 2 dep on nr of points
            self.winddim = min(2,len(self.lat))

    def addpoint(self,lat,lon,winddir,windspd,windalt=None):
        """ addpoint: adds a lat,lon position with a wind direction [deg]
                                                     and wind speedd [m/s]
            Optionally an array with altitudes can be used in which case windspd
            and wind speed need to have the same dimension
        """

        # If scalar, copy into table for altitude axis
        if not(type(windalt) in [ndarray,list]) and windalt == None: # scalar to array
            prof3D = False # no wind profile, just one value
            wspd   = ones(self.nalt)*windspd
            wdir   = ones(self.nalt)*winddir
            vnaxis = wspd*cos(radians(wdir)+pi)
            veaxis = wspd*sin(radians(wdir)+pi)

        # if list or array, convert to alt axis of wind field
        else:
            prof3D = True # switch on 3D parameter as an altitude array is given
            wspd   = array(windspd)
            wdir   = array(winddir)
            altvn  = wspd*cos(radians(wdir)+pi)
            altve  = wspd*sin(radians(wdir)+pi)
            alttab = windalt

            vnaxis = interp(self.altaxis, alttab, altvn)
            veaxis = interp(self.altaxis, alttab, altve)

#        print array([vnaxis]).transpose()
        self.lat    = append(self.lat,lat)
        self.lon    = append(self.lon,lon)

        idx = len(self.lat)-1

        if self.nvec==0:
            self.vnorth = array([vnaxis]).transpose()
            self.veast  = array([veaxis]).transpose()

        else:
            self.vnorth = append(self.vnorth,array([vnaxis]).transpose(),axis=1)
            self.veast  = append(self.veast, array([veaxis]).transpose(),axis=1)

        if self.winddim<3: # No 3D => set dim to 0,1 or 2 dep on nr of points
            self.winddim = min(2,len(self.lat))

        if prof3D:
            self.winddim = 3
            self.iprof.append(idx)

        self.nvec = self.nvec+1

        return idx # return index of added point
    
    def getdata(self,userlat,userlon,useralt=0.0): # in case no altitude specified and field is 3D, use sea level wind
        eps = 1e-20 # [m2] to avoid divison by zero for using exact same points

        swvector = (type(userlat)==list or type(userlat)==ndarray)
        if swvector:
            npos = len(userlat)
        else:
            npos = 1
        # Convert user input to right shape: columns for positions
        lat = array(userlat).reshape((1,npos))
        lon = array(userlon).reshape((1,npos))

        # Make altitude into an array, with zero or float value broadcast over npos
        if type(useralt)==ndarray:
            alt = useralt
        elif type(useralt)==list:
            alt = array(useralt)
        elif type(useralt)==float:
            alt = useralt*ones(npos)
        else:
            alt = zeros(npos)

        # Check if RGI functions are present, if so use them for interpolation
        if self.fe is not None and self.fn is not None:
            vnorth = self.fn(concatenate((alt.reshape(1,-1), lat, lon), axis=0).T)
            veast  = self.fe(concatenate((alt.reshape(1,-1), lat, lon), axis=0).T)
        else:
            # Check dimension of wind field
            if self.winddim == 0:   # None = no wind
                vnorth = zeros(npos)
                veast  = zeros(npos)
    
            elif self.winddim == 1: # Constant = one point defined, so constant wind
                vnorth = ones(npos)*self.vnorth[0,0]
                veast  = ones(npos)*self.veast[0,0]
    
            elif self.winddim >= 2: # 2D/3D field = more points defined but no altitude profile
    
                #---- Get horizontal weight factors
    
                # Average cosine for flat-eartyh approximation
                cavelat = cos(radians(0.5*(lat+array([self.lat]).transpose())))
    
                # Lat and lon distance in 60 nm units (1 lat degree)
                dy = lat - array([self.lat]).transpose() #(nvec,npos)
                dx = cavelat*(lon - array([self.lon]).transpose())
    
                # Calulate invesre distance squared
                invd2   = 1./(eps+dx*dx+dy*dy) # inverse of distance squared
    
                # Normalize weights
                sumsid2 = ones((1,self.nvec)).dot(invd2) # totals to normalize weights
                totals = repeat(sumsid2,self.nvec,axis=0) # scale up dims to (nvec,npos)
                
                horfact = invd2/totals # rows x col = nvec x npos, weight factors
    
                #---- Altitude interpolation
    
                # No altitude profiles used: do 2D planar interpolation only
                if self.winddim == 2 or ((type(useralt) not in (list,ndarray)) and useralt==0.0): # 2D field no altitude interpolation
                    vnorth  = self.vnorth[0,:].dot(horfact)
                    veast   = self.veast[0,:].dot(horfact)
    
                # 3D interpolation as one or more points contain altitude profile
                else:
    
                    # Get altitude index as float for alt interpolation
                    idxalt = maximum(0., minimum(self.altaxis[-1]-eps, alt) / self.altstep) # find right index
    
                    # Convert to index and factor
                    ialt   = floor(idxalt).astype(int) # index array for lower altitude
                    falt   = idxalt-ialt  # factor for upper value
    
                    # Altitude interpolation combined with horizontal
                    nvec   = len(self.lon) # Get number of definition points
    
                    # North wind (y-direction ot lat direction)
                    vn0    = (self.vnorth[ialt,:]*horfact.T).dot(ones((nvec,1))) # hor interpolate lower alt (npos x)
                    vn1    = (self.vnorth[ialt+1,:]*horfact.T).dot(ones((nvec,1))) # hor interpolate lower alts (npos x)
                    vnorth = (1.-falt)*(vn0.reshape(npos)) + falt*(vn1.reshape(npos)) # As 1D array
    
                    # East wind (x-direction or lon direction)
                    ve0    = (self.veast[ialt,:]*horfact.T).dot(ones((nvec,1)))
                    ve1    = (self.veast[ialt+1,:]*horfact.T).dot(ones((nvec,1)))
                    veast  = (1.-falt)*(ve0.reshape(npos)) + falt*(ve1.reshape(npos)) # As 1D array

        # Return same type as positons were given
        if type(userlat)==ndarray:
            return vnorth,veast

        elif type(userlat)==list:
            return list(vnorth),list(veast)

        else:
            return float(vnorth),float(veast)

    def remove(self,idx): # remove a point using the returned index when it was added
        if idx<len(self.lat):
            self.lat = delete(self.lat,idx)
            self.lon = delete(self.lat,idx)

            self.vnorth = delete(self.vnorth,idx,axis=1)
            self.veast  = delete(self.veast ,idx,axis=1)

            if idx in self.iprof:
                self.iprof.remove(idx)

            if self.winddim<3 or len(self.iprof)==0 or len(self.lat)==0:
                self.winddim = min(2,len(self.lat)) # Check for 0, 1D, 2D or 3D

        return



''' BlueSky: The open-source ATM simulator.'''
import importlib
# from bluesky import settings, stack
from bluesky.pathfinder import resource


__all__ = ['settings', 'stack', 'tools', 'cmdargs']

# Constants
BS_OK = 0
BS_ARGERR = 1
BS_FUNERR = 2
BS_CMDERR = 4

# simulation states
INIT, HOLD, OP, END = list(range(4))

# Startup flags
mode = ''
gui = ''

# Main singleton objects in BlueSky
ref = None
net = None
traf = None
navdb = None
sim = None
scr = None
server = None


def init(mode='sim', configfile=None, scenfile=None, discoverable=False,
         gui=None, detached=False, workdir=None, group_id=None, **kwargs):
    ''' Initialize bluesky modules.

        Arguments:
        - mode: Running mode of this bluesky process [sim/client/server]
        - configfile: Load a different configuration file [filename]
        - scenfile: Start with a running scenario [filename]
        - discoverable: Make server discoverable through UDP (only relevant 
          when this process is running a server) [True/False]
        - gui: Gui type (only when mode is client or server) [qtgl/pygame/console]
        - detached: Run with or without networking (only when mode is sim) [True/False]
        - workdir: Pass a custom working directory (instead of cwd or ~/bluesky)
        - group_id: Explicitly set (part of) the connection identifier string.
                    Server does this when spawning a node
    '''

    # Argument checking
    assert mode in ('sim', 'client', 'server'), f'BlueSky init: Unrecognised mode {mode}. '\
        'Possible modes are sim, client, and server.'
    assert gui in (None, 'qtgl', 'pygame', 'console'), f'BlueSky init: Unrecognised gui type {gui}. '\
        'Possible types are qtgl, pygame, and console.'
    if discoverable:
        assert mode == 'server', 'BlueSky init: Discoverable can only be set in server mode.'
    if scenfile:
        assert mode != 'client', 'BlueSky init: Scenario file cannot be passed to a client.'
    if gui:
        assert mode != 'sim' or gui == 'pygame', 'BlueSky init: Gui type shouldn\'t be specified in sim mode.'
    if detached:
        assert mode == 'sim', 'BlueSky init: Detached operation is only available in sim mode.'

    # Keep track of mode and gui type.
    globals()['mode'] = mode
    globals()['gui'] = gui

    # Initialise resource localisation, and set custom working directory if present
    from bluesky import pathfinder
    pathfinder.init(workdir)

    # Initialize global settings, possibly loading a custom config file
    from bluesky import settings
    settings.init(configfile)

    from bluesky import stack, tools

    # Initialise tools
    from bluesky import tools
    tools.init()

    # Initialise reference data object
    from bluesky import refdata
    globals()['ref'] = refdata.RefData()

    # Load navdatabase in all versions of BlueSky
    # Only the headless server doesn't need this
    if mode == "sim" or gui is not None:
        from bluesky.navdatabase import Navdatabase
        global navdb
        navdb = Navdatabase()

    # If mode is server-gui or server-headless start the networking server
    if mode == 'server':
        global server
        from bluesky.network.server import Server
        server = Server(discoverable, configfile, scenfile)

    # The remaining objects are only instantiated in the sim nodes
    if mode == 'sim':
        from bluesky.traffic import Traffic
        from bluesky.simulation import Simulation
        if gui == 'pygame':
            from bluesky.ui.pygame import Screen
            from bluesky.network.detached import Node
        else:
            from bluesky.simulation import ScreenIO as Screen
            if detached:
                from bluesky.network.detached import Node
            else:
                from bluesky.network.node import Node

        from bluesky.core import varexplorer

        # Initialize singletons
        global traf, sim, scr, net
        traf = Traffic()
        sim = Simulation()
        scr = Screen()
        net = Node(group_id)

        # Initialize remaining modules
        varexplorer.init()

    from bluesky.core import plugin
    plugin.init(mode)
    from bluesky import stack
    stack.init(mode)
    if scenfile and mode == 'sim':
        stack.stack(f'IC {scenfile}')


def __getattr__(name):
    if name in __all__:
        return importlib.import_module(f'.{name}', __name__)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from collections.abc import Collection
import numpy as np
import matplotlib
matplotlib.rcParams['font.size'] = 5

from matplotlib.figure import Figure
from matplotlib.backend_bases import key_press_handler
import matplotlib.pyplot as plt

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QTabWidget, QVBoxLayout, QScrollArea, QWidget
matplotlib.use('QtAgg')

from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas#, \
    #NavigationToolbar2QT as NavigationToolbar

from bluesky import stack
from bluesky.core import Entity
from bluesky.network import subscriber, context as ctx


class InfoWindow(Entity):
    ''' Top-level window containing simulation information such as plots. '''
    def __init__(self):
        super().__init__()
        self.view = QTabWidget()
        self.view.setDocumentMode(True)
        self.view.resize(600, 500)
        self.plottab = None
        self.statetab = StateTab(self.view)
        self.view.addTab(self.statetab, 'Simulation state')

    @stack.command(name="INFO")
    def show(self) -> bool:
        self.view.show()
        return True

    # def keyPressEvent(self, event):
    #     if event.key() == Qt.Key.Key_Escape:
    #         self.view.close()

    def add_plot_tab(self):
        self.plottab = PlotTab()
        self.view.addTab(self.plottab, 'Graphs')

    @subscriber
    def plot(self, **data):
        if not self.plottab:
            self.add_plot_tab()

        # A show flag is sent each time the PLOT command is called
        if data.pop('show', False):
            self.show()

        # A reset flag is sent upon sim reset to indicate removal of sim plots
        if data.pop('reset', False):
            self.plottab.remove_plots(ctx.sender_id)

        self.plottab.update_plots(data, ctx.sender_id)


class StateTab(QScrollArea):
    def __init__(self, parent):
        super().__init__(parent)
        container = QWidget()
        self.setWidget(container)
        self.setWidgetResizable(True)

class PlotTab(QScrollArea):
    ''' InfoWindow tab for plots. '''
    def __init__(self):
        super().__init__()
        self.layout = QVBoxLayout()
        container = QWidget()
        container.setLayout(self.layout)
        self.layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setWidget(container)
        self.setWidgetResizable(True)
        self.plots = dict()

    def remove_plots(self, sender_id):
        ''' Remove plots from reset simulation from plot window. '''
        for (sender_plt, fig), plot in dict(self.plots).items():
            if sender_id == sender_plt:
                self.layout.removeWidget(plot)
                self.plots.pop((sender_plt, fig))

    def update_plots(self, data, sender):
        ''' Update plots in this tab using incoming data. '''
        for fig, figdata in data.items():
            # First extract plot data when present
            x, y = figdata.pop('x', None), figdata.pop('y', None)

            plot = self.plots.get((sender, fig))
            if not plot:
                # If plot doesn't exist yet, create it
                plot = Plot(self, **figdata)
                self.plots[(sender, fig)] = plot
                self.layout.addWidget(plot)
            elif figdata:
                # When passed, apply updated figure settings
                plot.set(**figdata)

            if x is not None and y is not None:
                plot.update_data(x, y)

class Plot(FigureCanvas):
    def __init__(self, parent, plot_type='line', **kwargs):
        super().__init__(plt.figure())
        self.setParent(parent)
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.setFocus()
        self.setFixedHeight(350)
        self.axes = self.figure.add_subplot(111, **kwargs)
        self.figure.tight_layout(pad=8)
        self.plots = []
        self.plot_type = plot_type
        self.data = []

    def set(self, **kwargs):
        for flag, value in kwargs.items():
            if flag == 'legend':
                if len(self.plots) < len(value):
                    for _ in range(len(value) - len(self.plots)):
                        if self.plot_type == 'line':
                            lineobj = self.axes.plot(np.array([]), np.array([]))[0]
                        self.plots.append(lineobj)
                self.axes.legend(value)

    def update_data(self, xdata, ydata):
        if self.plot_type == 'line':
            if isinstance(xdata, Collection) and not isinstance(ydata, Collection):
                ydata = [ydata] * len(xdata)
            elif not isinstance(xdata, Collection) and isinstance(ydata, Collection):
                xdata = [xdata] * len(ydata)
            elif not isinstance(xdata, Collection) and not isinstance(ydata, Collection):
                xdata = [xdata]
                ydata = [ydata]
            npoints = len(xdata)
            if len(self.plots) < npoints:
                for _ in range(npoints - len(self.plots)):
                    lineobj = self.axes.plot(np.array([]), np.array([]))[0]
                    self.plots.append(lineobj)

            for p, x, y in zip(self.plots, xdata, ydata):
                p.set_xdata(np.append(p.get_xdata(), x))
                p.set_ydata(np.append(p.get_ydata(), y))
                p.axes.relim()
                p.axes.autoscale_view()

        elif self.plot_type == 'boxplot' and len(ydata):
            nnewplots = len(ydata) - len(self.data)
            if nnewplots > 0:
                self.data.extend(nnewplots * [[]])
            for i, d in enumerate(ydata):
                self.data[i].extend(d)
            # self.data = [x.extend(n) for x, n in zip(self.data, ydata)]
            self.data = [d for d in self.data if d]
            if len(self.data):
                self.axes.cla()
                self.axes.boxplot(self.data)


        self.draw()
        self.flush_events()
